from typing import List, Mapping, Optional
from core.messagebus.messagebus import Admin
from kafka.admin import KafkaAdminClient, NewTopic
from kafka import __version__
import asyncio
import logging


class KafkaEventConfig():
    def __init__(
        self,
        topic_name: str,
        num_partitions: int = 1,
        replication_factor: int = 1
    ):
        self.topic_name = topic_name
        self.num_partitions = num_partitions
        self.replication_factor = replication_factor


class KafkaAdmin(Admin):
    def __init__(
        self,
        logger: logging.Logger,
        bootstrap_servers: str,
        configs: Mapping[str, KafkaEventConfig],
        client_id: str = 'kafka-admin-' + __version__,
        security_protocol='PLAINTEXT',
        sasl_mechanism="PLAIN",
        sasl_plain_password=None,
        sasl_plain_username=None,
        sasl_kerberos_service_name='kafka',
        sasl_kerberos_domain_name=None,
        sasl_oauth_token_provider=None,
    ):
        self.logger = logger
        self.configs = configs
        self.bootstrap_servers = bootstrap_servers
        self.client_id = client_id
        self.security_protocol = security_protocol
        self.sasl_mechanism = sasl_mechanism
        self.sasl_plain_password = sasl_plain_password
        self.sasl_plain_username = sasl_plain_username
        self.sasl_kerberos_service_name = sasl_kerberos_service_name
        self.sasl_kerberos_domain_name = sasl_kerberos_domain_name
        self.sasl_oauth_token_provider = sasl_oauth_token_provider
        self._existing_events: Mapping[str, bool] = {}

    async def create_events(self, event_names: List[str]):
        # Only handle non-existing events
        event_names = [
            event_name for event_name in event_names
            if event_name not in self._existing_events
        ]
        if len(event_names) == 0:
            return
        # Create topics
        topics = [
            self.get_new_topic(event_name) for event_name in event_names
        ]
        try:
            admin_client = self._create_connection()
            admin_client.create_topics(topics)
            admin_client.close()
        except (asyncio.CancelledError, GeneratorExit, Exception):
            self.logger.error(
                ' '.join([
                    '🐼 [kafka-admin] Something wrong when ',
                    f'creating topics: {topics}'
                ]),
                exc_info=True
            )
        for event_name in event_names:
            self._existing_events[event_name] = True

    async def delete_events(self, event_names: List[str]):
        # Only handle existing events
        event_names = [
            event_name for event_name in event_names
            if event_name in self._existing_events
        ]
        if len(event_names) == 0:
            return
        # Create topic names
        topic_names = [
            self.get_topic_name(event_name) for event_name in event_names
            if event_name in self._existing_events
        ]
        try:
            admin_client = self._create_connection()
            admin_client.delete_topics(topic_names)
            admin_client.close()
            for event_name in event_names:
                del self._existing_events[event_name]
        except (asyncio.CancelledError, GeneratorExit, Exception):
            self.logger.error(
                ' '.join([
                    '🐼 [kafka-admin] Something wrong when ',
                    f'deleting topics: {topic_names}'
                ]),
                exc_info=True
            )

    def get_config(self, event_name: str) -> KafkaEventConfig:
        if event_name in self.configs:
            return self.configs[event_name]
        return KafkaEventConfig(topic_name=event_name)

    def get_topic_name(self, event_name: str) -> str:
        event_config = self.get_config(event_name)
        return event_config.topic_name

    def get_new_topic(self, event_name: str) -> NewTopic:
        event_config = self.get_config(event_name)
        topic = NewTopic(
            name=event_config.topic_name,
            num_partitions=event_config.num_partitions,
            replication_factor=event_config.replication_factor
        )
        return topic

    def _create_connection(self) -> KafkaAdminClient:
        return KafkaAdminClient(
            bootstrap_servers=self.bootstrap_servers,
            client_id=self.client_id,
            security_protocol=self.security_protocol,
            sasl_mechanism=self.sasl_mechanism,
            sasl_plain_password=self.sasl_plain_password,
            sasl_plain_username=self.sasl_plain_username,
            sasl_kerberos_service_name=self.sasl_kerberos_service_name,
            sasl_kerberos_domain_name=self.sasl_kerberos_domain_name,
            sasl_oauth_token_provider=self.sasl_oauth_token_provider
        )


def must_get_kafka_admin(
    logger: logging.Logger,
    kafka_admin: Optional[KafkaAdmin],
    bootstrap_servers: str,
    client_id='kafka-admin-' + __version__,
    security_protocol='PLAINTEXT',
    sasl_mechanism="PLAIN",
    sasl_plain_password=None,
    sasl_plain_username=None,
    sasl_kerberos_service_name='kafka',
    sasl_kerberos_domain_name=None,
    sasl_oauth_token_provider=None,
) -> KafkaAdmin:
    if kafka_admin is None:
        return KafkaAdmin(
            logger=logger,
            bootstrap_servers=bootstrap_servers,
            configs={},
            client_id=client_id,
            security_protocol=security_protocol,
            sasl_mechanism=sasl_mechanism,
            sasl_plain_password=sasl_plain_password,
            sasl_plain_username=sasl_plain_username,
            sasl_kerberos_service_name=sasl_kerberos_service_name,
            sasl_kerberos_domain_name=sasl_kerberos_domain_name,
            sasl_oauth_token_provider=sasl_oauth_token_provider,
        )
    return kafka_admin



from zrb.helper.typecheck import typechecked
from zrb.helper.typing import Any, List, Mapping


@typechecked
def fetch_compose_file_env_map(data: Any) -> Mapping[str, str]:
    global_env_dict = {}
    if 'services' not in data:
        return global_env_dict
    for service in data['services']:
        if 'environment' not in data['services'][service]:
            continue
        environments = data['services'][service]['environment']
        if isinstance(environments, list):
            for environment in environments:
                parts: List[str] = environment.split('=')
                if len(parts) > 0:
                    env_str = str(parts[1])
                    env_dict = parse_compose_file_env_string(env_str)
                    global_env_dict.update(env_dict)
        if isinstance(environments, dict):
            for _, env_str in environments.items():
                env_str = str(env_str)
                env_dict = parse_compose_file_env_string(env_str)
                global_env_dict.update(env_dict)
    return global_env_dict


@typechecked
def parse_compose_file_env_string(env_str: str) -> Mapping[str, str]:
    env_dict = {}
    stack = []
    key = ''
    value = ''
    index = 0
    while index < len(env_str):
        char = env_str[index]
        if char == '{':
            stack.append(index)
        elif char == '}':
            start = stack.pop()
            if not stack:
                segment = env_str[start+1:index]
                if ':-' in segment:
                    key, value = segment.split(':-', 1)
                    if value.startswith('${') and value.endswith('}'):
                        sub_dict = parse_compose_file_env_string(value)
                        env_dict.update(sub_dict)
                else:
                    key = segment
                    value = ''
                if '${' in value:
                    value = ''
                env_dict[key] = value
                key = ''
                value = ''
        index += 1
    return env_dict



from zrb import Env
from ._config import DEPLOYMENT_DIR, MODULES
import jsons

app_enable_otel_env = Env(
    name='APP_ENABLE_OTEL',
    default='{{ "1" if input.enable_snake_zrb_app_name_monitoring else "0" }}'
)

image_env = Env(
    name='IMAGE',
    os_name='CONTAINER_ZRB_ENV_PREFIX_IMAGE',
    default='{{input.snake_zrb_app_name_image}}'
)

pulumi_backend_url_env = Env(
    name='PULUMI_BACKEND_URL',
    os_name='PULUMI_ZRB_ENV_PREFIX_BACKEND_URL',
    default=f'file://{DEPLOYMENT_DIR}/state'
)

pulumi_config_passphrase_env = Env(
    name='PULUMI_CONFIG_PASSPHRASE',
    os_name='PULUMI_ZRB_ENV_PREFIX_CONFIG_PASSPHRASE',
    default='secret'
)

deployment_modules_env = Env(
    name='MODULES',
    os_name='DEPLOYMENT_CONFIG_ZRB_ENV_PREFIX_MODULES',
    default=jsons.dumps(MODULES)
)

deployment_mode_env = Env(
    name='MODE',
    os_name='DEPLOYMENT_CONFIG_ZRB_ENV_PREFIX_MODE',
    default='{{input.snake_zrb_app_name_deploy_mode}}'
)

deployment_enable_monitoring_env = Env(
    name='ENABLE_MONITORING',
    os_name='DEPLOYMENT_CONFIG_ZRB_ENV_PREFIX_ENABLE_MONITORING',
    default='{{ 1 if input.enable_snake_zrb_app_name_monitoring else 0 }}'
)


from zrb.helper.typing import Optional, JinjaTemplate
from zrb.helper.typecheck import typechecked
from zrb.task_env.constant import RESERVED_ENV_NAMES
from zrb.helper.string.modification import double_quote
import os

# flake8: noqa E501

@typechecked
class Env():
    '''
    Env Represents an environment configuration for a task, encapsulating details such as environment name, OS-specific 
    environment name, default values, and rendering behavior.

    Attributes:
        name (str): Environment name as recognized by the task.
        os_name (Optional[str]): The corresponding name in the OS's environment, if different from 'name'. You can set os_name to empty string if you don't want the environment to be linked to OS environment name.
        default (JinjaTemplate): Default value of the environment variable.
        should_render (bool): Flag to determine if the environment value should be rendered.
    
    Examples:
        >>> from zrb import Env, Task
        >>> task = Task(
        >>>     name='task',
        >>>     envs=[
        >>>         Env(name='DATABASE_URL', os_name='SYSTEM_DATABASE_URL', default='postgresql://...')
        >>>     ]
        >>> )
    '''

    def __init__(
        self,
        name: str,
        os_name: Optional[str] = None,
        default: JinjaTemplate = '',
        should_render: bool = True,
    ):
        if name in RESERVED_ENV_NAMES:
            raise ValueError(f'Forbidden input name: {name}')
        self.__name: str = name
        self.__os_name: str = os_name if os_name is not None else name
        self.__default: str = default
        self.__should_render: bool = should_render

    def get_name(self) -> str:
        '''
        Retrieves the name of the environment variable.

        Returns:
            str: The name of the environment variable.
        '''
        return self.__name

    def get_os_name(self) -> Optional[str]:
        '''
        Retrieves the OS-specific name of the environment variable.

        Returns:
            Optional[str]: The OS-specific name of the environment variable.
        '''
        return self.__os_name

    def get_default(self) -> str:
        '''
        Retrieves the default value of the environment variable.

        Returns:
            str: The default value of the environment variable.
        '''
        return self.__default

    def should_render(self) -> bool:
        '''
        Determines whether the environment value should be rendered.

        Returns:
            bool: True if the environment value should be rendered, False otherwise.
        '''
        return self.__should_render

    def get(self, prefix: str = '') -> str:
        '''
        Retrieves the value of the environment variable, considering an optional prefix.

        Args:
            prefix (str): An optional prefix to distinguish different environments (e.g., 'DEV', 'PROD').

        Returns:
            str: The value of the environment variable, considering the prefix and default value.

        Examples:
            >>> from zrb import Env
            >>> import os
            >>> os.environ['DEV_SERVER'] = 'localhost'
            >>> os.environ['PROD_SERVER'] = 'example.com'
            >>> env = Env(name='HOST', os_name='SERVER', default='0.0.0.0')
            >>> print(env.get('DEV'))   # will show 'localhost'
            >>> print(env.get('PROD'))  # will show 'example.com'
            >>> print(env.get('STAG'))  # will show '0.0.0.0'
        '''
        if self.__os_name == '':
            return self.__default
        prefixed_name = self.__get_prefixed_name(self.__os_name, prefix)
        if prefixed_name in os.environ and os.environ[prefixed_name] != '':
            return os.environ[prefixed_name]
        if self.__os_name in os.environ and os.environ[self.__os_name] != '':
            return os.environ[self.__os_name]
        return self.__default

    def __get_prefixed_name(self, name: str, prefix: str):
        '''
        Constructs the prefixed name of the environment variable.

        This method is intended for internal use only. 

        Args:
            name (str): The base name of the environment variable.
            prefix (str): The prefix to be added to the name.

        Returns:
            str: The prefixed name of the environment variable.
        '''
        if prefix is None or prefix == '':
            return name
        return prefix + '_' + name

    def __repr__(self) -> str:
        cls_name = self.__class__.__name__
        name = double_quote(self.__name)
        os_name = 'None' if self.__os_name is None else double_quote(self.__os_name)
        default = double_quote(self.__default)
        return f'<{cls_name} {name} os_name={os_name} default={default}>'



from zrb import EnvFile
from ._config import (
    RESOURCE_DIR, APP_TEMPLATE_ENV_FILE_NAME, DEPLOYMENT_TEMPLATE_ENV_FILE_NAME
)
import os

app_env_file = EnvFile(
    path=APP_TEMPLATE_ENV_FILE_NAME, prefix='ZRB_ENV_PREFIX'
)

compose_env_file = EnvFile(
    path=os.path.join(RESOURCE_DIR, 'docker-compose.env'),
    prefix='CONTAINER_ZRB_ENV_PREFIX'
)

deployment_app_env_file = EnvFile(
    path=APP_TEMPLATE_ENV_FILE_NAME,
    prefix='DEPLOYMENT_APP_ZRB_ENV_PREFIX'
)

deployment_config_env_file = EnvFile(
    path=DEPLOYMENT_TEMPLATE_ENV_FILE_NAME,
    prefix='DEPLOYMENT_CONFIG_ZRB_ENV_PREFIX'
)


from zrb.helper.typing import Any
from zrb.task.decorator import python_task
from zrb.runner import runner
from zrb.config.config import version

###############################################################################
# Task Definitions
###############################################################################


@python_task(
    name='version',
    description='Get Zrb version',
    runner=runner
)
async def get_version(*args: Any, **kwargs: Any) -> str:
    return version



from zrb.helper.typing import (
    Any, Callable, Iterable, Optional, Union, TypeVar
)
from zrb.helper.typecheck import typechecked
from zrb.task.checker import Checker
from zrb.task.any_task import AnyTask
from zrb.task.any_task_event_handler import (
    OnTriggered, OnWaiting, OnSkipped, OnStarted, OnReady, OnRetry, OnFailed
)
from zrb.task_env.env import Env
from zrb.task_env.env_file import EnvFile
from zrb.task_group.group import Group
from zrb.task_input.any_input import AnyInput

import croniter
import datetime

TTimeWatcher = TypeVar('TTimeWatcher', bound='TimeWatcher')


@typechecked
class TimeWatcher(Checker):
    '''
    TimeWatcher will wait for any changes specified on  path.

    Once the changes detected, TimeWatcher will be completed
    and <task-name>.scheduled-time xcom will be set.
    '''

    def __init__(
        self,
        name: str = 'watch-path',
        group: Optional[Group] = None,
        inputs: Iterable[AnyInput] = [],
        envs: Iterable[Env] = [],
        env_files: Iterable[EnvFile] = [],
        icon: Optional[str] = None,
        color: Optional[str] = None,
        description: str = '',
        upstreams: Iterable[AnyTask] = [],
        on_triggered: Optional[OnTriggered] = None,
        on_waiting: Optional[OnWaiting] = None,
        on_skipped: Optional[OnSkipped] = None,
        on_started: Optional[OnStarted] = None,
        on_ready: Optional[OnReady] = None,
        on_retry: Optional[OnRetry] = None,
        on_failed: Optional[OnFailed] = None,
        schedule: str = '',
        checking_interval: Union[int, float] = 1,
        progress_interval: Union[int, float] = 30,
        should_execute: Union[bool, str, Callable[..., bool]] = True
    ):
        Checker.__init__(
            self,
            name=name,
            group=group,
            inputs=inputs,
            envs=envs,
            env_files=env_files,
            icon=icon,
            color=color,
            description=description,
            upstreams=upstreams,
            on_triggered=on_triggered,
            on_waiting=on_waiting,
            on_skipped=on_skipped,
            on_started=on_started,
            on_ready=on_ready,
            on_retry=on_retry,
            on_failed=on_failed,
            checking_interval=checking_interval,
            progress_interval=progress_interval,
            should_execute=should_execute,
        )
        self._schedule = schedule
        self._scheduled_time: Optional[datetime.datetime] = None
        self._rendered_schedule: str = ''

    def copy(self) -> TTimeWatcher:
        return super().copy()

    def to_function(
        self,
        env_prefix: str = '',
        raise_error: bool = True,
        is_async: bool = False,
        show_done_info: bool = True
    ) -> Callable[..., bool]:
        return super().to_function(
            env_prefix, raise_error, is_async, show_done_info
        )

    async def run(self, *args: Any, **kwargs: Any) -> bool:
        self._rendered_schedule = self.render_str(self._schedule)
        margin = datetime.timedelta(seconds=0.001)
        slightly_before_check_time = datetime.datetime.now() - margin
        cron = croniter.croniter(
            self._rendered_schedule, slightly_before_check_time
        )
        self._scheduled_time = cron.get_next(datetime.datetime)
        self.set_task_xcom(key='scheduled-time', value=self._scheduled_time)
        return await super().run(*args, **kwargs)

    async def inspect(self, *args: Any, **kwargs: Any) -> bool:
        label = f'Watching {self._rendered_schedule}'
        now = datetime.datetime.now()
        if now > self._scheduled_time:
            self.print_out_dark(
                f'{label} (Meet {self._scheduled_time})'
            )
            return True
        self.show_progress(
            f'{label} (Waiting for {self._scheduled_time})'
        )
        return False



from zrb import BoolInput, ChoiceInput, StrInput
import os

enable_monitoring_input = BoolInput(
    name='enable-kebab-zrb-app-name-monitoring',
    description='Enable "kebab-zrb-app-name" monitoring',
    prompt='Enable "kebab-zrb-app-name" monitoring?',
    default=False
)

local_input = BoolInput(
    name='local-kebab-zrb-app-name',
    description='Use "kebab-zrb-app-name" on local machine',
    prompt='Use "kebab-zrb-app-name" on local machine?',
    default=True
)

run_mode_input = ChoiceInput(
    name='kebab-zrb-app-name-run-mode',
    description='"kebab-zrb-app-name" run mode (monolith/microservices)',
    prompt='Run "kebab-zrb-app-name" as a monolith or microservices?',
    choices=['monolith', 'microservices'],
    default='monolith'
)

https_input = BoolInput(
    name='kebab-zrb-app-name-https',
    description='Whether "kebab-zrb-app-name" run on HTTPS',
    prompt='Is "kebab-zrb-app-name" run on HTTPS?',
    default=False
)

host_input = StrInput(
    name='kebab-zrb-app-name-host',
    description='Hostname of "kebab-zrb-app-name"',
    prompt='Hostname of "kebab-zrb-app-name"',
    default='localhost'
)

image_input = StrInput(
    name='kebab-zrb-app-name-image',
    description='Image name of "kebab-zrb-app-name"',
    prompt='Image name of "kebab-zrb-app-name"',
    default='zrb-app-image-name:latest'
)

deploy_mode_input = ChoiceInput(
    name='kebab-zrb-app-name-deploy-mode',
    description='"kebab-zrb-app-name" deploy mode (monolith/microservices)',
    prompt='Deploy "kebab-zrb-app-name" as a monolith or microservices?',
    choices=['monolith', 'microservices'],
    default='monolith'
)

pulumi_stack_input = StrInput(
    name='kebab-zrb-app-name-pulumi-stack',
    description='Pulumi stack name for "kebab-zrb-app-name"',
    prompt='Pulumi stack name for "kebab-zrb-app-name"',
    default=os.getenv('ZRB_ENV', 'dev')
)


from zrb.helper.typing import Any, Optional, Union
from zrb.helper.typecheck import typechecked

from zrb.task_input.base_input import BaseInput

# flake8: noqa E501

@typechecked
class StrInput(BaseInput):
    '''
    A specialized input class for handling string-based inputs in various tasks.

    `StrInput` extends `BaseInput` to manage string-type inputs, supporting features like 
    default values, prompts, flags, and other customization options. This class is useful 
    for tasks requiring textual input, such as names, descriptions, or any other string parameters.

    Args:
        name (str): The name of the input, used as an identifier.
        shortcut (Optional[str]): An optional shortcut string for the input.
        default (Optional[Any]): The default value for the input, expected to be a string if set.
        description (Optional[str]): A brief description of the input's purpose.
        show_default (Union[bool, str, None]): Option to display the default value. Can be a boolean or string.
        prompt (Union[bool, str]): A boolean or string to prompt the user for input. If `True`, uses the default prompt.
        confirmation_prompt (Union[bool, str]): If `True`, the user is asked to confirm the input.
        prompt_required (bool): If `True’, the prompt for input is mandatory.
        hide_input (bool): If `True’, hides the input value, typically used for sensitive data.
        is_flag (Optional[bool]): Indicates if the input is a flag. If `True’, the input accepts boolean flag values.
        flag_value (Optional[Any]): The value associated with the flag if `is_flag` is `True`.
        multiple (bool): If `True’, allows multiple string values for the input.
        count (bool): If `True’, counts the occurrences of the input.
        allow_from_autoenv (bool): If `True’, enables automatic population of the input from environment variables.
        hidden (bool): If `True’, keeps the input hidden in help messages or documentation.
        show_choices (bool): If `True’, shows any restricted choices for the input value.
        show_envvar (bool): If `True’, displays the associated environment variable, if applicable.
        nargs (int): The number of arguments that the input can accept.
        should_render (bool): If `True’, renders the input in the user interface or command-line interface.

    Examples:
        >>> str_input = StrInput(name='username', default='user123', description='Enter your username')
        >>> str_input.get_default()
        'user123'
    '''

    def __init__(
        self,
        name: str,
        shortcut: Optional[str] = None,
        default: Optional[Any] = None,
        description: Optional[str] = None,
        show_default: Union[bool, str, None] = None,
        prompt: Union[bool, str] = True,
        confirmation_prompt: Union[bool, str] = False,
        prompt_required: bool = True,
        hide_input: bool = False,
        is_flag: Optional[bool] = None,
        flag_value: Optional[Any] = None,
        multiple: bool = False,
        count: bool = False,
        allow_from_autoenv: bool = True,
        hidden: bool = False,
        show_choices: bool = True,
        show_envvar: bool = False,
        nargs: int = 1,
        should_render: bool = True
    ):
        BaseInput.__init__(
            self,
            name=name,
            shortcut=shortcut,
            default=default,
            description=description,
            show_default=show_default,
            prompt=prompt,
            confirmation_prompt=confirmation_prompt,
            prompt_required=prompt_required,
            hide_input=hide_input,
            is_flag=is_flag,
            flag_value=flag_value,
            multiple=multiple,
            count=count,
            allow_from_autoenv=allow_from_autoenv,
            type=str,
            hidden=hidden,
            show_choices=show_choices,
            show_envvar=show_envvar,
            nargs=nargs,
            should_render=should_render
        )



from zrb import CmdTask, DockerComposeTask, Task, Env, runner
from zrb.builtin.group import project_group
from ._config import CURRENT_DIR, APP_DIR, RESOURCE_DIR
from ._helper import (
    activate_support_compose_profile, should_start_support_container,
    should_start_local_microservices, should_start_local_monolith
)
from ._checker import (
    rabbitmq_checker, rabbitmq_management_checker, redpanda_console_checker,
    kafka_outside_checker, kafka_plaintext_checker, pandaproxy_outside_checker,
    pandaproxy_plaintext_checker, app_local_checker
)
from ._input import (
    local_input, https_input, run_mode_input, enable_monitoring_input,
    host_input, image_input
)
from ._env import app_enable_otel_env
from ._env_file import app_env_file
from ._get_start_microservices import get_start_microservices
from .frontend import build_snake_zrb_app_name_frontend
from .container import remove_snake_zrb_app_name_container, compose_env_file
import os


###############################################################################
# ⚙️ init-kebab-zrb-task-name-support-container
###############################################################################

init_snake_zrb_app_name_support_container = DockerComposeTask(
    icon='🔥',
    name='init-kebab-zrb-app-name-support-container',
    inputs=[
        local_input,
        host_input,
        image_input,
    ],
    should_execute=should_start_support_container,
    upstreams=[
        remove_snake_zrb_app_name_container
    ],
    cwd=RESOURCE_DIR,
    setup_cmd=activate_support_compose_profile,
    compose_cmd='up',
    compose_flags=['-d'],
    compose_env_prefix='CONTAINER_ZRB_ENV_PREFIX',
    env_files=[compose_env_file]
)

###############################################################################
# ⚙️ start-kebab-zrb-task-name-support-container
###############################################################################

start_snake_zrb_app_name_support_container = DockerComposeTask(
    icon='🐳',
    name='start-kebab-zrb-app-name-support-container',
    description='Start human readable zrb app name container',
    inputs=[
        local_input,
        enable_monitoring_input,
        host_input,
        https_input,
        image_input,
    ],
    should_execute=should_start_support_container,
    upstreams=[init_snake_zrb_app_name_support_container],
    cwd=RESOURCE_DIR,
    setup_cmd=activate_support_compose_profile,
    compose_cmd='logs',
    compose_flags=['-f'],
    compose_env_prefix='CONTAINER_ZRB_ENV_PREFIX',
    env_files=[compose_env_file],
    checkers=[
        rabbitmq_checker,
        rabbitmq_management_checker,
        kafka_outside_checker,
        kafka_plaintext_checker,
        redpanda_console_checker,
        pandaproxy_outside_checker,
        pandaproxy_plaintext_checker,
    ]
)

###############################################################################
# ⚙️ prepare-kebab-zrb-task-name-backend
###############################################################################

prepare_snake_zrb_app_name_backend = CmdTask(
    icon='🚤',
    name='prepare-kebab-zrb-app-name-backend',
    description='Prepare backend for human readable zrb app name',
    group=project_group,
    cwd=APP_DIR,
    cmd_path=[
        os.path.join(CURRENT_DIR, 'cmd', 'activate-venv.sh'),
        os.path.join(CURRENT_DIR, 'cmd', 'app-prepare-backend.sh'),
    ]
)
runner.register(prepare_snake_zrb_app_name_backend)

###############################################################################
# ⚙️ start-monolith-kebab-zrb-task-name
###############################################################################

start_monolith_snake_zrb_app_name = CmdTask(
    icon='🚤',
    name='start-monolith-kebab-zrb-app-name',
    inputs=[
        local_input,
        enable_monitoring_input,
        run_mode_input,
        host_input,
        https_input
    ],
    should_execute=should_start_local_monolith,
    upstreams=[
        start_snake_zrb_app_name_support_container,
        build_snake_zrb_app_name_frontend,
        prepare_snake_zrb_app_name_backend,
    ],
    cwd=APP_DIR,
    env_files=[app_env_file],
    envs=[app_enable_otel_env],
    cmd_path=[
        os.path.join(CURRENT_DIR, 'cmd', 'activate-venv.sh'),
        os.path.join(CURRENT_DIR, 'cmd', 'app-start.sh'),
    ],
    checkers=[
        app_local_checker,
    ]
)

###############################################################################
# ⚙️ start-kebab-zrb-task-name-gateway
###############################################################################

start_snake_zrb_app_name_gateway = CmdTask(
    icon='🚪',
    name='start-kebab-zrb-app-name-gateway',
    inputs=[
        local_input,
        enable_monitoring_input,
        run_mode_input,
        host_input,
        https_input
    ],
    should_execute=should_start_local_microservices,
    upstreams=[
        start_snake_zrb_app_name_support_container,
        build_snake_zrb_app_name_frontend,
        prepare_snake_zrb_app_name_backend,
    ],
    cwd=APP_DIR,
    env_files=[
        app_env_file,
    ],
    envs=[
        Env(name='APP_DB_AUTO_MIGRATE', default='false', os_name=''),
        Env(name='APP_ENABLE_EVENT_HANDLER', default='false', os_name=''),
        Env(name='APP_ENABLE_RPC_SERVER', default='false', os_name=''),
        app_enable_otel_env,
    ],
    cmd_path=[
        os.path.join(CURRENT_DIR, 'cmd', 'activate-venv.sh'),
        os.path.join(CURRENT_DIR, 'cmd', 'app-start.sh'),
    ],
    checkers=[
        app_local_checker,
    ]
)

###############################################################################
# ⚙️ start-kebab-zrb-task-name
###############################################################################

start_snake_zrb_app_name = Task(
    icon='🚤',
    name='start-kebab-zrb-app-name',
    description='Start human readable zrb app name',
    group=project_group,
    upstreams=[
        start_monolith_snake_zrb_app_name,
        start_snake_zrb_app_name_gateway,
        *get_start_microservices(
            upstreams=[
                start_snake_zrb_app_name_support_container,
                build_snake_zrb_app_name_frontend,
                prepare_snake_zrb_app_name_backend,
            ]
        )
    ],
    run=lambda *args, **kwargs: kwargs.get('_task').print_out('🆗')
)
runner.register(start_snake_zrb_app_name)



from zrb import DockerComposeTask, runner
from zrb.builtin.group import project_group
from ._config import RESOURCE_DIR
from ._input import local_input, image_input
from ._env import image_env

###############################################################################
# ⚙️ build-kebab-zrb-task-name-image
###############################################################################

build_snake_zrb_app_name_image = DockerComposeTask(
    icon='🏭',
    name='build-kebab-zrb-app-name-image',
    description='Build human readable zrb app name image',
    group=project_group,
    inputs=[
        local_input,
        image_input,
    ],
    envs=[image_env],
    should_execute='{{ input.local_snake_zrb_app_name}}',
    cwd=RESOURCE_DIR,
    compose_cmd='build',
    compose_args=[
        'kebab-zrb-app-name'
    ],
    compose_env_prefix='CONTAINER_ZRB_ENV_PREFIX',
)
runner.register(build_snake_zrb_app_name_image)

###############################################################################
# ⚙️ push-kebab-zrb-task-name-image
###############################################################################

push_snake_zrb_app_name_image = DockerComposeTask(
    icon='📰',
    name='push-kebab-zrb-app-name-image',
    description='Push human readable zrb app name image',
    group=project_group,
    inputs=[
        local_input,
        image_input,
    ],
    envs=[image_env],
    upstreams=[build_snake_zrb_app_name_image],
    cwd=RESOURCE_DIR,
    compose_cmd='push',
    compose_args=[
        'kebab-zrb-app-name'
    ],
    compose_env_prefix='CONTAINER_ZRB_ENV_PREFIX',
)
runner.register(push_snake_zrb_app_name_image)



from zrb.helper.docstring import get_markdown_from_docstring


class SampleClass:
    """
    This is a sample class for testing.

    Attributes:
        attribute1 (int): An example attribute.
    """

    def method1(self, param1: str) -> bool:
        """
        An example method.

        Args:
            param1 (str): A sample parameter.

        Returns:
            bool: The return value.

        Examples:
            >>> sc = SampleClass()
            >>> sc.method1("test")
        """
        return True


def test_get_markdown_from_docstring():
    expected_output = (
        "## `SampleClass`\n\n"
        "This is a sample class for testing.\n\n"
        "__Attributes:__\n\n"
        "- `attribute1` (`int`): An example attribute.\n"
        "\n### `SampleClass.method1`\n\n"
        "An example method.\n\n"
        "__Arguments:__\n\n"
        "- `param1` (`str`): A sample parameter.\n\n"
        "__Returns:__\n\n"
        "`bool`: The return value.\n\n"
        "__Examples:__\n\n"
        "```python\n"
        "sc = SampleClass()\n"
        "sc.method1(\"test\")\n"
        "```\n\n"
    )
    assert get_markdown_from_docstring(SampleClass) == expected_output



from typing import Any, Callable
import inspect


async def run_async(
    fn: Callable, *args: Any, **kwargs: Any
) -> Any:
    if inspect.iscoroutinefunction(fn):
        return await fn(*args, **kwargs)
    return fn(*args, **kwargs)



from typing import Any, Mapping
from zrb import (
    runner, python_task, Task, CmdTask, DockerComposeTask, FlowTask, Checker,
    ResourceMaker, RsyncTask, RemoteCmdTask, PathChecker, PathWatcher,
    TimeWatcher, PortChecker, RecurringTask, HTTPChecker, BaseRemoteCmdTask,
    Notifier, Env, EnvFile, Group, Input, BoolInput, ChoiceInput, FloatInput,
    IntInput, PasswordInput, StrInput
)
from zrb.helper.docstring import get_markdown_from_docstring

import os
import re
import subprocess
import sys
import time
import tomli

CURRENT_DIR = os.path.dirname(__file__)
PLAYGROUND_DIR = os.path.join(CURRENT_DIR, 'playground')
IS_PLAYGROUND_EXIST = os.path.isdir(PLAYGROUND_DIR)

if IS_PLAYGROUND_EXIST:
    sys.path.append(PLAYGROUND_DIR)
    from playground import zrb_init
    assert zrb_init


with open(os.path.join(CURRENT_DIR, 'pyproject.toml'), 'rb') as f:
    toml_dict = tomli.load(f)
    VERSION = toml_dict['project']['version']


def inject_doc(markdown_file_name: str, cls):
    docstring_markdown = get_markdown_from_docstring(cls)
    with open(markdown_file_name, 'r') as file:
        original_content = file.read()
    pattern = r'<!--start-doc-->.*?<!--end-doc-->'
    replacement_text = '\n'.join([
        '<!--start-doc-->',
        docstring_markdown,
        '<!--end-doc-->',
    ])
    new_content = re.sub(
        pattern, replacement_text, original_content, flags=re.DOTALL
    )
    with open(markdown_file_name, 'w') as file:
        file.write(new_content)


###############################################################################
# Group Definitions
###############################################################################

playground_group = Group(
    name='playground', description='Playground related tasks'
)

###############################################################################
# 🔤 Input Definitions
###############################################################################

zrb_version_input = StrInput(
    name='zrb-version',
    description='Zrb version',
    prompt='Zrb version',
    default=VERSION
)

zrb_image_name_input = StrInput(
    name='zrb-image-name',
    description='Zrb image name',
    prompt='Zrb image name',
    default=f'docker.io/stalchmst/zrb:{VERSION}'
)

zrb_latest_image_name_input = StrInput(
    name='zrb-latest-image-name',
    description='Zrb latest image name',
    prompt='Zrb latest image name',
    default='docker.io/stalchmst/zrb:latest'
)

build_zrb_input = BoolInput(name='build-zrb', default=True)
install_symlink_input = BoolInput(name='install-symlink', default=True)
create_playground_input = BoolInput(
    name='create-playground', default=not IS_PLAYGROUND_EXIST
)

###############################################################################
# 🌱 Env Definitions
###############################################################################

zrb_version_env = Env(
    name='ZRB_VERSION',
    os_name='',
    default='{{input.zrb_version}}'
)

zrb_image_env = Env(
    name='ZRB_IMAGE',
    os_name='',
    default='{{input.zrb_image_name}}'
)

zrb_latest_image_env = Env(
    name='ZRB_IMAGE',
    os_name='',
    default='{{input.zrb_latest_image_name}}'
)

###############################################################################
# ⚙️ make-docs
###############################################################################


@python_task(
    name='make-docs',
    description='Make documentation',
    runner=runner
)
def make_docs(*args: Any, **kwargs: Any):
    task: Task = kwargs.get('_task')
    doc_dir = os.path.join(CURRENT_DIR, 'docs')
    doc_concept_dir = os.path.join(doc_dir, 'concepts')
    doc_concept_task_dir = os.path.join(doc_concept_dir, 'task')
    doc_concept_task_input_dir = os.path.join(doc_concept_dir, 'task-input')
    configs: Mapping[str, Any] = {
        os.path.join(doc_concept_dir, 'task-group.md'): Group,
        os.path.join(doc_concept_dir, 'task-env.md'): Env,
        os.path.join(doc_concept_dir, 'task-env-file.md'): EnvFile,
        os.path.join(doc_concept_task_dir, 'README.md'): Task,
        os.path.join(doc_concept_task_dir, 'base-remote-cmd-task.md'): BaseRemoteCmdTask,  # noqa
        os.path.join(doc_concept_task_dir, 'checker.md'): Checker,
        os.path.join(doc_concept_task_dir, 'cmd-task.md'): CmdTask,
        os.path.join(doc_concept_task_dir, 'docker-compose-task.md'): DockerComposeTask,  # noqa
        os.path.join(doc_concept_task_dir, 'flow-task.md'): FlowTask,
        os.path.join(doc_concept_task_dir, 'http-checker.md'): HTTPChecker,
        os.path.join(doc_concept_task_dir, 'path-checker.md'): PathChecker,
        os.path.join(doc_concept_task_dir, 'path-watcher.md'): PathWatcher,
        os.path.join(doc_concept_task_dir, 'port-checker.md'): PortChecker,
        os.path.join(doc_concept_task_dir, 'python-task.md'): python_task,
        os.path.join(doc_concept_task_dir, 'recurring-task.md'): RecurringTask,
        os.path.join(doc_concept_task_dir, 'remote-cmd-task.md'): RemoteCmdTask,  # noqa
        os.path.join(doc_concept_task_dir, 'resource-maker.md'): ResourceMaker,
        os.path.join(doc_concept_task_dir, 'rsync-task.md'): RsyncTask,
        os.path.join(doc_concept_task_dir, 'notifier.md'): Notifier,
        os.path.join(doc_concept_task_dir, 'time-watcher.md'): TimeWatcher,
        os.path.join(doc_concept_task_input_dir, 'README.md'): Input,
        os.path.join(doc_concept_task_input_dir, 'bool-input.md'): BoolInput,
        os.path.join(doc_concept_task_input_dir, 'choice-input.md'): ChoiceInput,  # noqa
        os.path.join(doc_concept_task_input_dir, 'float-input.md'): FloatInput,
        os.path.join(doc_concept_task_input_dir, 'int-input.md'): IntInput,
        os.path.join(doc_concept_task_input_dir, 'password-input.md'): PasswordInput,  # noqa
        os.path.join(doc_concept_task_input_dir, 'str-input.md'): StrInput,
    }
    for file_name, cls in configs.items():
        task.print_out(f'Inject `{cls.__name__} docstring` to {file_name}')
        inject_doc(file_name, cls)


###############################################################################
# ⚙️ show-trigger-info
###############################################################################

@python_task(
    name='show-trigger-info'
)
def show_trigger_info(*args: Any, **kwargs: Any):
    task = kwargs.get('_task')
    file = task.get_xcom('watch-path.file')
    if file != '':
        task.print_out(f'Trigger: {file}')


###############################################################################
# ⚙️ remake-docs
###############################################################################

remake_docs = CmdTask(
    name='remake-docs',
    cwd=CURRENT_DIR,
    envs=[
        Env('ZRB_SHOW_TIME', os_name='', default='0')
    ],
    upstreams=[show_trigger_info],
    cmd='zrb make-docs'
)

###############################################################################
# ⚙️ auto-make-docs
###############################################################################

auto_make_docs = RecurringTask(
    name='auto-make-docs',
    description='Make documentation whenever there is any changes in the code',
    triggers=[
        PathWatcher(
            path=os.path.join(CURRENT_DIR, 'src', 'zrb', '**', '*.py'),
        )
    ],
    task=remake_docs,
    single_execution=True
)
runner.register(auto_make_docs)

###############################################################################
# ⚙️ build
###############################################################################

build = CmdTask(
    name='build',
    description='Build Zrb',
    upstreams=[make_docs],
    cwd=CURRENT_DIR,
    cmd=[
        'set -e',
        'echo "🤖 Build zrb distribution"',
        f'rm -Rf {CURRENT_DIR}/dist',
        'git add . -A',
        'flit build',
    ],
)
skippable_build: CmdTask = build.copy()
skippable_build.add_input(build_zrb_input)
skippable_build.set_should_execute('{{ input.build_zrb}}')
runner.register(build)

###############################################################################
# ⚙️ publish-pip
###############################################################################

publish_pip = CmdTask(
    name='publish-pip',
    description='Publish zrb to pypi',
    upstreams=[build],
    cwd=CURRENT_DIR,
    cmd=[
        'set -e',
        'echo "🤖 Publish zrb to pypi"',
        'flit publish --repository pypi',
    ]
)
runner.register(publish_pip)

###############################################################################
# ⚙️ publish-pip-test
###############################################################################

publish_pip_test = CmdTask(
    name='publish-pip-test',
    description='Publish zrb to testpypi',
    upstreams=[build],
    cwd=CURRENT_DIR,
    cmd=[
        'set -e',
        'echo "🤖 Publish zrb to testpypi"',
        'flit publish --repository testpypi',
    ]
)
runner.register(publish_pip_test)

###############################################################################
# ⚙️ check-pip
###############################################################################

check_pip = HTTPChecker(
    name='check-pip',
    inputs=[zrb_version_input],
    is_https=True,
    host='pypi.org',
    url='pypi/zrb/{{ input.zrb_version }}/json',
    port=443
)

###############################################################################
# ⚙️ build-image
###############################################################################

build_image = DockerComposeTask(
    name='build-image',
    description='Build docker image',
    upstreams=[check_pip],
    inputs=[
        zrb_version_input,
        zrb_image_name_input,
    ],
    envs=[
        zrb_version_env,
        zrb_image_env,
    ],
    cwd=f'{CURRENT_DIR}/docker',
    compose_cmd='build',
    compose_args=['zrb']
)
runner.register(build_image)

###############################################################################
# ⚙️ build-latest-image
###############################################################################

build_latest_image = DockerComposeTask(
    name='build-latest-image',
    description='Build docker image',
    upstreams=[
        check_pip,
        build_image,
    ],
    inputs=[
        zrb_version_input,
        zrb_latest_image_name_input,
    ],
    envs=[
        zrb_version_env,
        zrb_latest_image_env,
    ],
    cwd=f'{CURRENT_DIR}/docker',
    compose_cmd='build',
    compose_args=['zrb']
)
runner.register(build_latest_image)

###############################################################################
# ⚙️ stop-container
###############################################################################

stop_container = DockerComposeTask(
    name='stop-container',
    description='remove docker container',
    inputs=[
        zrb_image_name_input
    ],
    envs=[
        zrb_image_env
    ],
    cwd=f'{CURRENT_DIR}/docker',
    compose_cmd='down'
)
runner.register(stop_container)

###############################################################################
# ⚙️ start-container
###############################################################################

start_container = DockerComposeTask(
    name='start-container',
    description='Run docker container',
    upstreams=[
        build_image,
        stop_container
    ],
    inputs=[
        zrb_version_input,
        zrb_image_name_input,
    ],
    envs=[
        zrb_version_env,
        zrb_image_env,
    ],
    cwd=f'{CURRENT_DIR}/docker',
    compose_cmd='up',
    compose_flags=['-d']
)
runner.register(start_container)

###############################################################################
# ⚙️ push-image
###############################################################################

push_image = DockerComposeTask(
    name='push-image',
    description='Push docker image',
    upstreams=[build_image],
    inputs=[
        zrb_version_input,
        zrb_image_name_input,
    ],
    envs=[
        zrb_version_env,
        zrb_image_env,
    ],
    cwd=f'{CURRENT_DIR}/docker',
    compose_cmd='push',
    compose_args=['zrb']
)
runner.register(push_image)

###############################################################################
# ⚙️ push-latest-image
###############################################################################

push_latest_image = DockerComposeTask(
    name='push-latest-image',
    description='Push docker image',
    upstreams=[
        build_latest_image,
        push_image,
    ],
    inputs=[
        zrb_version_input,
        zrb_latest_image_name_input,
    ],
    envs=[
        zrb_version_env,
        zrb_latest_image_env,
    ],
    cwd=f'{CURRENT_DIR}/docker',
    compose_cmd='push',
    compose_args=['zrb']
)
runner.register(push_latest_image)

###############################################################################
# ⚙️ publish
###############################################################################

publish = FlowTask(
    name='publish',
    description='Publish new version',
    steps=[
        publish_pip,
        push_latest_image,
    ]
)
runner.register(publish)

###############################################################################
# ⚙️ install-symlink
###############################################################################

install_symlink = CmdTask(
    name='install-symlink',
    description='Install Zrb as symlink',
    upstreams=[skippable_build],
    cmd=[
        'set -e',
        f'cd {CURRENT_DIR}',
        'echo "🤖 Install zrb"',
        'flit install --symlink',
    ]
)
skippable_install_symlink: CmdTask = install_symlink.copy()
skippable_install_symlink.add_input(
    build_zrb_input,
    install_symlink_input
)
skippable_install_symlink.set_should_execute('{{ input.install_symlink}}')
runner.register(install_symlink)

###############################################################################
# ⚙️ test
###############################################################################

test_only = CmdTask(
    name='test',
    description='Run zrb test',
    inputs=[
        StrInput(
            name='test',
            shortcut='t',
            description='Specific test case (i.e., test/file.py::test_name)',
            prompt='Test (i.e., test/file.py::test_name)',
            default=''
        ),
    ],
    cmd=[
        'set -e',
        'echo "🤖 Perform test"',
        'pytest -vv \\',
        '  --ignore-glob="**/template/**/test" \\',
        '  --ignore-glob="**/generator/**/app"  \\',
        '  --ignore=playground \\',
        '  --cov=zrb \\',
        '  --cov-config=".coveragerc" \\',
        '  --cov-report=html \\',
        '  --cov-report=term \\',
        '  --cov-report=term-missing {{input.test}}'
    ],
    retry=0,
    checking_interval=1
)

test = test_only.copy()
test.add_upstream(install_symlink)
runner.register(test)

###############################################################################
# ⚙️ re-test
###############################################################################
retest = test_only.copy()
retest.add_upstream(show_trigger_info)

###############################################################################
# ⚙️ auto-test
###############################################################################

auto_test = RecurringTask(
    name='auto-test',
    description='Run zrb test, automatically',
    upstreams=[test],
    task=retest,
    triggers=[
        PathWatcher(
            path=os.path.join(CURRENT_DIR, 'src', 'zrb', '**', '*.py'),
        ),
        PathWatcher(
            path=os.path.join(CURRENT_DIR, 'test', '**', '*.py'),
            ignored_path=[
                os.path.join('**', 'template', '**', 'test'),
                os.path.join('**', 'generator', '**', 'app'),
                os.path.join('**', 'resource_maker', 'app'),
            ]
        )
    ],
    single_execution=True
)
runner.register(auto_test)

###############################################################################
# ⚙️ serve-test
###############################################################################

serve_test = CmdTask(
    name='serve-test',
    description='Serve zrb test result',
    inputs=[
        StrInput(
            name='port',
            shortcut='p',
            description='Port to serve coverage result',
            prompt='Serve coverage on port',
            default='9000'
        )
    ],
    upstreams=[auto_test],
    cmd=[
        'set -e',
        'echo "🤖 Serve coverage report"',
        'python -m http.server {{input.port}} \\',
        f'  --directory "{CURRENT_DIR}/htmlcov"',
    ],
    checkers=[
        HTTPChecker(port='{{input.port}}')
    ],
    retry=0,
    checking_interval=0.3
)
runner.register(serve_test)

###############################################################################
# ⚙️ monitor
###############################################################################

monitor = Task(
    name='monitor',
    description='Monitor any changes and perform actions accordingly',
    upstreams=[serve_test, auto_make_docs]
)
runner.register(monitor)

###############################################################################
# ⚙️ playground create
###############################################################################

create_playground = CmdTask(
    name='create',
    description='Create playground',
    group=playground_group,
    upstreams=[skippable_install_symlink],
    cwd=CURRENT_DIR,
    cmd_path=os.path.join(CURRENT_DIR, 'playground-create.sh'),
    retry=0,
    preexec_fn=None
)
skippable_create_playground: CmdTask = create_playground.copy()
skippable_create_playground.add_input(
    build_zrb_input,
    install_symlink_input,
    create_playground_input
)
skippable_create_playground.set_should_execute('{{ input.create_playground}}')
runner.register(create_playground)

###############################################################################
# ⚙️ playground test-fastapp
###############################################################################

test_fastapp_playground = CmdTask(
    name='test-fastapp',
    description='Test Fastapp',
    group=playground_group,
    upstreams=[skippable_create_playground],
    cwd=CURRENT_DIR,
    cmd_path=[
        os.path.join(CURRENT_DIR, 'playground-init.sh'),
        os.path.join(CURRENT_DIR, 'playground-test-fastapp.sh')
    ],
    retry=0,
    preexec_fn=None
)
runner.register(test_fastapp_playground)

###############################################################################
# ⚙️ playground test-install-symlink
###############################################################################

test_install_playground_symlink = CmdTask(
    name='test-install-symlink',
    description='Test installing symlink',
    group=playground_group,
    upstreams=[skippable_create_playground],
    cwd=CURRENT_DIR,
    cmd_path=[
        os.path.join(CURRENT_DIR, 'playground-init.sh'),
        os.path.join(CURRENT_DIR, 'playground-test-install-symlink.sh')
    ],
    retry=0,
    preexec_fn=None
)
runner.register(test_install_playground_symlink)

###############################################################################
# ⚙️ playground test
###############################################################################

test_playground = CmdTask(
    name='test',
    description='Test playground',
    group=playground_group,
    upstreams=[
        test_fastapp_playground,
        test_install_playground_symlink
    ],
    cmd='echo Test performed',
    retry=0
)
runner.register(test_playground)


###############################################################################
# ⚙️ prepare-profile
###############################################################################

prepare_profile = CmdTask(
    name='prepare-profile',
    description='Prepare profile',
    cmd='python -m cProfile -o .cprofile.prof $(pwd)/src/zrb/__main__.py',
    # cmd='python -m cProfile -o .cprofile.prof -m zrb',
    retry=0
)

###############################################################################
# ⚙️ profile
###############################################################################

profile = CmdTask(
    name='profile',
    description='Visualize profile',
    upstreams=[prepare_profile],
    cmd='flameprof .cprofile.prof > .cprofile.svg',
    retry=0
)
runner.register(profile)

###############################################################################
# ⚙️ benchmark-import
###############################################################################


@python_task(
    name='benchmark',
    description='Benchmark',
    runner=runner
)
def benchmark(*args: Any, **kwargs: Any):
    statements = [
        'import zrb.runner',
        'import zrb.task.decorator',
        'import zrb.task.any_task',
        'import zrb.task.any_task_event_handler',
        'import zrb.task.parallel',
        'import zrb.task.task',
        'import zrb.task.cmd_task',
        'import zrb.task.docker_compose_task',
        'import zrb.task.base_remote_cmd_task',
        'import zrb.task.remote_cmd_task',
        'import zrb.task.rsync_task',
        'import zrb.task.checker',
        'import zrb.task.http_checker',
        'import zrb.task.port_checker',
        'import zrb.task.path_checker',
        'import zrb.task.path_watcher',
        'import zrb.task.time_watcher',
        'import zrb.task.resource_maker',
        'import zrb.task.flow_task',
        'import zrb.task.recurring_task',
        'import zrb.task_input.any_input',
        'import zrb.task_input.task_input',
        'import zrb.task_input.bool_input',
        'import zrb.task_input.choice_input',
        'import zrb.task_input.float_input',
        'import zrb.task_input.int_input',
        'import zrb.task_input.password_input',
        'import zrb.task_input.str_input',
        'import zrb.task_env.env',
        'import zrb.task_env.env_file',
        'import zrb.task_group.group',
        'import zrb.helper.default_env',
        'from zrb import builtin',
        'import zrb',
    ]
    results = []
    for statement in statements:
        start_time = time.time()
        subprocess.run(['python', '-c', statement])
        end_time = time.time()
        results.append(
            f'[{end_time - start_time:.5f} seconds] {statement}'
        )
    return '\n'.join(results)



from config import (
    zrb_app_name, app_broker_type, app_rmq_connection_string,
    app_kafka_bootstrap_servers, app_kafka_security_protocol,
    app_kafka_sasl_mechanism, app_kafka_sasl_user, app_kafka_sasl_pass
)
from core.messagebus import (
    Admin, Consumer, MessageSerializer, Publisher,
    MockAdmin, MockConsumer, MockPublisher,
    KafkaAdmin, KafkaConsumer, KafkaPublisher,
    RMQAdmin, RMQConsumer, RMQPublisher
)
from component.log import logger


def init_admin(
    default_admin: Admin
) -> Admin:
    if app_broker_type == 'rabbitmq':
        return RMQAdmin(
            logger=logger,
            connection_string=app_rmq_connection_string,
            configs={}
        )
    if app_broker_type == 'kafka':
        return KafkaAdmin(
            logger=logger,
            bootstrap_servers=app_kafka_bootstrap_servers,
            security_protocol=app_kafka_security_protocol,
            sasl_mechanism=app_kafka_sasl_mechanism,
            sasl_plain_username=app_kafka_sasl_user,
            sasl_plain_password=app_kafka_sasl_pass,
            configs={}
        )
    return default_admin


def init_publisher(
    serializer: MessageSerializer,
    admin: Admin,
    default_publisher: Publisher
) -> Publisher:
    if app_broker_type == 'rabbitmq':
        return RMQPublisher(
            logger=logger,
            connection_string=app_rmq_connection_string,
            serializer=serializer,
            rmq_admin=admin
        )
    if app_broker_type == 'kafka':
        return KafkaPublisher(
            logger=logger,
            bootstrap_servers=app_kafka_bootstrap_servers,
            security_protocol=app_kafka_security_protocol,
            sasl_mechanism=app_kafka_sasl_mechanism,
            sasl_plain_username=app_kafka_sasl_user,
            sasl_plain_password=app_kafka_sasl_pass,
            serializer=serializer,
            kafka_admin=admin
        )
    return default_publisher


def init_consumer(
    serializer: MessageSerializer,
    admin: Admin,
    default_consumer: Consumer
) -> Consumer:
    if app_broker_type == 'rabbitmq':
        return RMQConsumer(
            logger=logger,
            connection_string=app_rmq_connection_string,
            serializer=serializer,
            rmq_admin=admin
        )
    if app_broker_type == 'kafka':
        return KafkaConsumer(
            logger=logger,
            bootstrap_servers=app_kafka_bootstrap_servers,
            security_protocol=app_kafka_security_protocol,
            sasl_mechanism=app_kafka_sasl_mechanism,
            sasl_plain_username=app_kafka_sasl_user,
            sasl_plain_password=app_kafka_sasl_pass,
            group_id=zrb_app_name,
            serializer=serializer,
            kafka_admin=admin
        )
    return default_consumer


def init_message_serializer() -> MessageSerializer:
    # Add custom logic if necessary
    return MessageSerializer()


message_serializer = init_message_serializer()
mock_admin = MockAdmin()
mock_consumer = MockConsumer(logger=logger, serializer=message_serializer)
mock_publisher = MockPublisher(
    logger=logger, consumer=mock_consumer, serializer=message_serializer
)
admin = init_admin(default_admin=mock_admin)
consumer = init_consumer(
    serializer=message_serializer, admin=admin, default_consumer=mock_consumer
)
publisher = init_publisher(
    serializer=message_serializer,
    admin=admin,
    default_publisher=mock_publisher
)



from zrb.task.task import Task
from zrb.task.cmd_task import CmdTask


def test_task_execution_id_cannot_be_set_twice():
    task = Task(
        name='task',
        run=lambda *args, **kwargs: kwargs.get('_task').get_execution_id()
    )
    task._set_execution_id('execution_id_1')
    task._set_execution_id('execution_id_2')
    task._set_execution_id('execution_id_3')
    function = task.to_function()
    result = function()
    assert result == 'execution_id_1'


def test_consistent_task_upstream_execution_id():
    task_upstream_1 = Task(
        name='task-upstream-1',
        run=lambda *args, **kwargs: kwargs.get('_task').get_execution_id()
    )
    task_upstream_2 = CmdTask(
        name='task-upstream-2',
        cmd='echo $_ZRB_EXECUTION_ID'
    )
    task = Task(
        name='task',
        upstreams=[
            task_upstream_1, task_upstream_2
        ],
        return_upstream_result=True
    )
    function = task.to_function()
    result = function()
    assert result[0] == result[1].output



from zrb.task_group.group import Group
from zrb.task.task import Task


def test_init_simple_group():
    group = Group(name='group')
    assert group._get_full_cli_name() == 'group'
    assert group.get_description() == ''
    assert len(group.get_children()) == 0
    assert group.get_parent() is None
    assert str(group) == '<Group "group">'


def test_init_simple_group_with_description():
    group = Group(name='group', description='description')
    assert group._get_full_cli_name() == 'group'
    assert group.get_description() == 'description'
    assert len(group.get_children()) == 0
    assert group.get_parent() is None


def test_init_group_with_parent():
    grand_parent = Group(name='grand-parent')
    parent = Group(name='parent', parent=grand_parent)
    group = Group(name='group', parent=parent)
    assert group._get_full_cli_name() == 'grand-parent parent group'
    assert group.get_description() == ''
    assert len(group.get_children()) == 0
    assert group.get_parent()._get_full_cli_name() == 'grand-parent parent'
    assert str(group) == '<Group "grand-parent parent group">'


def test_group_with_children():
    group = Group(name='group')
    child_1 = Group(name='child-1', parent=group)
    child_2 = Group(name='child-2', parent=group)
    children = group.get_children()
    assert len(children) == 2
    assert children[0] == child_1
    assert children[1] == child_2
    assert group.get_parent() is None


def test_init_group_with_task():
    group = Group(name='group')
    task_1 = Task(name='task_1', group=group)
    task_2 = Task(name='task_2', group=group)
    tasks = group.get_tasks()
    assert len(tasks) == 2
    assert tasks[0] == task_1
    assert tasks[1] == task_2



from zrb.config.config import logging_level
from zrb.helper.typing import Any, Callable, List, Mapping, Union
from zrb.helper.typecheck import typechecked
from zrb.helper.log import logger
from zrb.helper.accessories.color import colored
from zrb.task_group.group import Group as TaskGroup
from zrb.task.any_task import AnyTask

import click
import logging
import sys

CliSubcommand = Union[click.Group, click.Command]


@typechecked
class Runner():
    '''
    Runner class.
    Any tasks registered to the runner will be accessible from the terminal
    '''

    def __init__(self, env_prefix: str = ''):
        if logging_level <= logging.INFO:
            logger.info(colored('Create runner', attrs=['dark']))
        self.__env_prefix = env_prefix
        self.__tasks: List[AnyTask] = []
        self.__registered_groups: Mapping[str, click.Group] = {}
        self.__top_levels: List[CliSubcommand] = []
        self.__subcommands: Mapping[str, List[click.Group]] = {}
        self.__registered_task_cli_name: List[str] = []
        if logging_level <= logging.INFO:
            logger.info(colored('Runner created', attrs=['dark']))

    def register(self, *tasks: AnyTask):
        for task in tasks:
            task._set_has_cli_interface()
            cli_name = task._get_full_cli_name()
            if cli_name in self.__registered_task_cli_name:
                raise RuntimeError(
                    f'Task "{cli_name}" has already been registered'
                )
            if logging_level <= logging.DEBUG:
                logger.debug(
                    colored(f'Register task: "{cli_name}"', attrs=['dark'])
                )
            self.__tasks.append(task)
            self.__registered_task_cli_name.append(cli_name)
            if logging_level <= logging.DEBUG:
                logger.debug(
                    colored(f'Task registered: "{cli_name}"', attrs=['dark'])
                )

    def serve(self, cli: click.Group) -> click.Group:
        for task in self.__tasks:
            subcommand = self.__create_cli_subcommand(task)
            if subcommand not in self.__top_levels:
                self.__top_levels.append(subcommand)
                cli.add_command(subcommand)
        return cli

    def __create_cli_subcommand(
        self, task: AnyTask
    ) -> Union[click.Group, click.Command]:
        subcommand: CliSubcommand = self.__create_cli_command(task)
        task_group = task._group
        while task_group is not None:
            group = self.__register_sub_command(task_group, subcommand)
            parent_group = task_group.get_parent()
            if parent_group is None:
                return group
            subcommand = group
            task_group = parent_group
        return subcommand

    def __register_sub_command(
        self, task_group: TaskGroup, subcommand: CliSubcommand
    ) -> click.Group:
        task_group_id = task_group._get_full_cli_name()
        group = self.__get_cli_group(task_group)
        if task_group_id not in self.__subcommands:
            self.__subcommands[task_group_id] = []
        if subcommand not in self.__subcommands[task_group_id]:
            group.add_command(subcommand)
            self.__subcommands[task_group_id].append(subcommand)
        return group

    def __get_cli_group(self, task_group: TaskGroup) -> click.Group:
        task_group_id = task_group._get_full_cli_name()
        if task_group_id in self.__registered_groups:
            return self.__registered_groups[task_group_id]
        group_cli_name = task_group.get_cli_name()
        group_description = task_group.get_description()
        group = click.Group(name=group_cli_name, help=group_description)
        self.__registered_groups[task_group_id] = group
        return group

    def __create_cli_command(self, task: AnyTask) -> click.Command:
        task_inputs = task._get_combined_inputs()
        task_cli_name = task.get_cli_name()
        task_description = task.get_description()
        task_function = task.to_function(
            env_prefix=self.__env_prefix, raise_error=True
        )
        callback = self.__wrap_task_function(task_function)
        command = click.Command(
            callback=callback, name=task_cli_name, help=task_description
        )
        # by default, add an argument named _args
        command.params.append(click.Argument(['_args'], nargs=-1))
        # add task inputs,
        # if there are inputs with the same name, choose the first.
        registered_input: Mapping[str, bool] = {}
        for task_input in task_inputs:
            if task_input.get_name() in registered_input:
                continue
            registered_input[task_input.get_name()] = True
            param_decl = task_input.get_param_decl()
            options = task_input.get_options()
            command.params.append(click.Option(param_decl, **options))
        return command

    def __wrap_task_function(
        self, function: Callable[..., Any]
    ) -> Callable[..., Any]:
        def wrapped_function(*args: Any, **kwargs: Any) -> Any:
            try:
                function(*args, **kwargs)
            except Exception:
                sys.exit(1)
        return wrapped_function



from zrb.helper.typing import List
from zrb.helper.typecheck import typechecked
from zrb.task.any_task import AnyTask
from zrb.task.cmd_task import CmdTask
from zrb.builtin.generator.common.task_input import (
    project_dir_input, app_name_input, module_name_input, entity_name_input,
)

import os

CURRENT_DIR = os.path.dirname(__file__)
CODEMOD_DIR = os.path.join(CURRENT_DIR, 'nodejs', 'codemod')

NAV_CONFIG_FILE_PATH = '{{input.project_dir}}/src/{{util.to_kebab_case(input.app_name)}}/src/frontend/src/lib/config/navData.ts' # noqa
ENV_VAR_NAME = 'navData'
NAV_TITLE = '{{util.to_pascal_case(input.entity_name)}}'
NAV_URL = '/{{util.to_kebab_case(input.module_name)}}/{{util.to_kebab_case(input.entity_name)}}' # noqa
NAV_PERMISSION = '{{util.to_snake_case(input.module_name)}}:{{util.to_snake_case(input.entity_name)}}:get' # noqa


@typechecked
def create_add_navigation_task(upstreams: List[AnyTask]) -> AnyTask:
    return CmdTask(
        name='add-navigation',
        inputs=[
            project_dir_input,
            app_name_input,
            module_name_input,
            entity_name_input,
        ],
        upstreams=upstreams,
        retry=0,
        cmd=f'node {CODEMOD_DIR}/dist/addNav.js "{NAV_CONFIG_FILE_PATH}" "{ENV_VAR_NAME}" "{NAV_TITLE}" "{NAV_URL}" "{NAV_PERMISSION}"' # noqa
    )



from zrb.helper.typing import Any, Callable, Iterable, Optional, Union
from zrb.helper.typecheck import typechecked
from zrb.task.base_task.base_task import BaseTask
from zrb.task.any_task import AnyTask
from zrb.task.any_task_event_handler import (
    OnTriggered, OnWaiting, OnSkipped, OnStarted, OnReady, OnRetry, OnFailed
)
from zrb.task_env.env import Env
from zrb.task_env.env_file import EnvFile
from zrb.task_group.group import Group
from zrb.task_input.any_input import AnyInput
from zrb.helper.accessories.icon import get_random_icon
from zrb.helper.string.modification import double_quote

import os
import subprocess

CURRENT_DIR = os.path.dirname(__file__)
NOTIFY_PS1_PATH = os.path.realpath(os.path.abspath(os.path.join(
    os.path.dirname(CURRENT_DIR), 'shell-scripts', 'notify.ps1'
)))


@typechecked
class Notifier(BaseTask):
    def __init__(
        self,
        name: str = 'port-check',
        group: Optional[Group] = None,
        inputs: Iterable[AnyInput] = [],
        envs: Iterable[Env] = [],
        env_files: Iterable[EnvFile] = [],
        icon: Optional[str] = None,
        color: Optional[str] = None,
        description: str = '',
        title: str = '',
        message: str = '',
        show_toast: bool = True,
        show_stdout: bool = True,
        upstreams: Iterable[AnyTask] = [],
        on_triggered: Optional[OnTriggered] = None,
        on_waiting: Optional[OnWaiting] = None,
        on_skipped: Optional[OnSkipped] = None,
        on_started: Optional[OnStarted] = None,
        on_ready: Optional[OnReady] = None,
        on_retry: Optional[OnRetry] = None,
        on_failed: Optional[OnFailed] = None,
        checking_interval: Union[int, float] = 0,
        retry: int = 2,
        retry_interval: Union[float, int] = 1,
        should_execute: Union[bool, str, Callable[..., bool]] = True
    ):
        BaseTask.__init__(
            self,
            name=name,
            group=group,
            inputs=inputs,
            envs=envs,
            env_files=env_files,
            icon=icon,
            color=color,
            description=description,
            upstreams=upstreams,
            on_triggered=on_triggered,
            on_waiting=on_waiting,
            on_skipped=on_skipped,
            on_started=on_started,
            on_ready=on_ready,
            on_retry=on_retry,
            on_failed=on_failed,
            checkers=[],
            checking_interval=checking_interval,
            retry=retry,
            retry_interval=retry_interval,
            should_execute=should_execute,
        )
        self._title = title if title != '' else name
        self._message = message if message != '' else get_random_icon()
        self._show_toast = show_toast
        self._show_stdout = show_stdout

    async def run(self, *args: Any, **kwargs: Any) -> str:
        title = self.render_str(self._title)
        message = self.render_str(self._message)
        notify_kwargs = {
            key: value
            for key, value in kwargs.items() if key not in ('title', 'message')
        }
        await self.notify(title, message, **notify_kwargs)
        return message

    async def notify(self, title: str, message: str, **kwargs: Any) -> None:
        task: BaseTask = kwargs.get('_task')
        if self._show_toast and _is_powershell_available():
            cmd = [
                'powershell.exe',
                '-ExecutionPolicy', 'Bypass',
                '-File', NOTIFY_PS1_PATH,
                '-Title', title,
                '-Message', message
            ]
            subprocess.run(cmd, stdout=subprocess.DEVNULL)
        if self._show_toast and _is_osascript_available():
            q_message = double_quote(message)
            q_title = double_quote(title)
            cmd = [
                'osascript',
                '-e',
                f'display notification "{q_message}" with title "{q_title}"'
            ]
        if self._show_toast and _is_notify_send_available():
            cmd = ['notify-send', title, message]
            subprocess.run(cmd, stdout=subprocess.DEVNULL)
        if self._show_stdout:
            task.print_out(message)
            task._play_bell()


def _is_powershell_available():
    try:
        subprocess.run(
            ['powershell.exe', '-Command', 'echo "Checking PowerShell"'],
            check=True, stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL
        )
        return True
    except (subprocess.CalledProcessError, FileNotFoundError, PermissionError):
        return False


def _is_notify_send_available():
    try:
        subprocess.run(
            ['notify-send', '--version'],
            check=True,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL
        )
        return True
    except (subprocess.CalledProcessError, FileNotFoundError, PermissionError):
        return False


def _is_osascript_available():
    try:
        subprocess.run(
            ['osascript', '-e', 'return'],
            check=True,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL
        )
        return True
    except (subprocess.CalledProcessError, FileNotFoundError, PermissionError):
        return False



from zrb.helper.typing import List, Optional
from zrb.helper.typecheck import typechecked
from dotenv import dotenv_values
from zrb.task_env.constant import RESERVED_ENV_NAMES
from zrb.helper.string.modification import double_quote
from zrb.task_env.env import Env

# flake8: noqa E501

@typechecked
class EnvFile():
    '''
    Represents a handler for an environment file, facilitating the creation and management of environment variables 
    (Env objects) based on the contents of the specified environment file.

    Attributes:
        path (str): The path to the environment file.
        prefix (Optional[str]): An optional prefix to be applied to environment variables.
        should_render (bool): Flag to determine if the environment values should be rendered.

    Examples:
        >>> from zrb import EnvFile, Task
        >>> import os
        >>> CURRENT_DIR = os.dirname(__file__)
        >>> task = Task(
        >>>     name='task',
        >>>     env_files=[
        >>>         EnvFile(path=os.path.join(CURRENT_DIR, '.env'), prefix='SYSTEM')
        >>>     ]
        >>> )
    '''

    def __init__(
        self,
        path: str,
        prefix: Optional[str] = None,
        should_render: bool = False
    ):
        self.__path = path
        self.__prefix = prefix.upper() if prefix is not None else None
        self.__should_render = should_render
        self.__env_list: List[Env] = []
        self.__env_list_fetched: bool = False

    def get_envs(self) -> List[Env]:
        '''
        Retrieves a list of Env objects based on the environment file. If a prefix is provided, it is 
        applied to the environment variable names.

        Returns:
            List[Env]: A list of Env objects representing the environment variables defined in the file.

        Examples:
            >>> from zrb import Env, EnvFile
            >>> env_file = EnvFile(path='some_file.env')
            >>> envs: List[Env] = env_file.get_envs()
        '''
        if self.__env_list_fetched:
            return self.__env_list
        env_list: List[Env] = []
        env_map = dotenv_values(self.__path)
        for key, value in env_map.items():
            if key in RESERVED_ENV_NAMES:
                continue
            os_name: Optional[str] = None
            if self.__prefix is not None and self.__prefix != '':
                os_name = f'{self.__prefix}_{key}'
            env_list.append(Env(
                name=key,
                os_name=os_name,
                default=value,
                should_render=self.__should_render
            ))
        self.__env_list = env_list
        self.__env_list_fetched = True
        return env_list

    def __repr__(self) -> str:
        cls_name = self.__class__.__name__
        path = double_quote(self.__path)
        prefix = double_quote(self.__prefix)
        return f'<{cls_name} path={path} prefix={prefix}>'



from zrb.task.any_task import AnyTask
from zrb.task.task import Task
from zrb.task.cmd_task import CmdTask
from zrb.task_input.str_input import StrInput
from zrb.task_env.env import Env
from zrb.task_env.env_file import EnvFile

import os

CURRENT_DIR = os.path.dirname(__file__)


def test_task_copy():
    def _run(*args, **kwargs) -> str:
        name = kwargs.get('name')
        task: AnyTask = kwargs.get('_task')
        env_map = task.get_env_map()
        environment = env_map.get('ENVIRONMENT')
        host = env_map.get('HOST')
        return f'hello {name} on {environment}, host: {host}'
    task = Task(
        name='task',
        inputs=[
            StrInput(name='name', default='Nicholas Flamel')
        ],
        envs=[
            Env(name='ENVIRONMENT', default='dev')
        ],
        env_files=[
            EnvFile(path=os.path.join(CURRENT_DIR, 'env_file.env'))
        ],
        run=_run,
        retry=0
    )
    new_task: Task = task.copy()
    new_task.set_name('new-task')
    assert task.get_cli_name() == 'task'
    assert new_task.get_cli_name() == 'new-task'
    assert task.get_description() == 'task'
    assert new_task.get_description() == 'new-task'
    new_task.set_description('new description')
    assert task.get_cli_name() == 'task'
    assert new_task.get_cli_name() == 'new-task'
    assert task.get_description() == 'task'
    assert new_task.get_description() == 'new description'
    new_task.set_icon('🔥')
    assert new_task.get_icon() == '🔥'
    new_task.set_color('green')
    assert new_task.get_color() == 'green'
    new_task.set_retry(1)
    assert new_task._retry == 1
    new_task.add_input(StrInput(name='name', default='Dumbledore'))
    new_task.add_env(Env(name='ENVIRONMENT', default='prod'))
    new_task.add_env_file(
        EnvFile(path=os.path.join(CURRENT_DIR, 'new_env_file.env'))
    )
    function = new_task.to_function()
    result = function()
    assert result == 'hello Dumbledore on prod, host: stalchmst.com'


def test_cmd_task_copy():
    task = CmdTask(
        name='task',
        inputs=[
            StrInput(name='name', default='Nicholas Flamel')
        ],
        envs=[
            Env(name='ENVIRONMENT', default='dev')
        ],
        env_files=[
            EnvFile(path=os.path.join(CURRENT_DIR, 'env_file.env'))
        ],
        cwd=CURRENT_DIR,
        cmd='echo hello $_INPUT_NAME on $ENVIRONMENT, host: $HOST',
        retry=0
    )
    new_task: CmdTask = task.copy()
    new_task.set_name('new-task')
    assert task.get_cli_name() == 'task'
    assert new_task.get_cli_name() == 'new-task'
    assert task.get_description() == 'task'
    assert new_task.get_description() == 'new-task'
    new_task.set_description('new description')
    assert task.get_cli_name() == 'task'
    assert new_task.get_cli_name() == 'new-task'
    assert task.get_description() == 'task'
    assert new_task.get_description() == 'new description'
    new_task.set_icon('🔥')
    assert new_task.get_icon() == '🔥'
    new_task.set_color('green')
    assert new_task.get_color() == 'green'
    new_task.set_retry(1)
    assert new_task._retry == 1
    new_task.set_cwd(os.path.dirname(CURRENT_DIR))
    assert task._cwd == CURRENT_DIR
    assert new_task._cwd == os.path.dirname(CURRENT_DIR)
    new_task.add_input(StrInput(name='name', default='Dumbledore'))
    new_task.add_env(Env(name='ENVIRONMENT', default='prod'))
    new_task.add_env_file(
        EnvFile(path=os.path.join(CURRENT_DIR, 'new_env_file.env'))
    )
    function = new_task.to_function()
    result = function()
    assert result.output == 'hello Dumbledore on prod, host: stalchmst.com'



from zrb.task.decorator import python_task
from zrb.task.cmd_task import CmdTask
from zrb.task.task import Task


def test_task_xcom():
    set_xcom_cmd = CmdTask(
        name='set-xcom-cmd',
        cmd='echo "hi{{task.set_xcom("one", "ichi")}}"'
    )

    @python_task(name='set-xcom-py')
    def set_xcom_py(*args, **kwargs):
        task: Task = kwargs.get('_task')
        task.set_xcom('two', 'ni')

    get_xcom_cmd = CmdTask(
        name='get-xcom-cmd',
        upstreams=[set_xcom_cmd, set_xcom_py],
        cmd=[
            'echo {{task.get_xcom("one")}}',
            'echo {{task.get_xcom("two")}}',
        ]
    )

    @python_task(
        name='get-xcom-py',
        upstreams=[set_xcom_cmd, set_xcom_py],
    )
    def get_xcom_py(*args, **kwargs):
        task: Task = kwargs.get('_task')
        return '\n'.join([
            task.get_xcom("one"),
            task.get_xcom("two"),
        ])

    test_xcom = Task(
        name='test-xcom',
        upstreams=[get_xcom_cmd, get_xcom_py],
        return_upstream_result=True
    )
    fn = test_xcom.to_function()
    result = fn()
    assert len(result) == 2
    assert result[0].output == 'ichi\nni'
    assert result[1] == 'ichi\nni'




from typing import Any, Mapping
from logging import Logger
from core.messagebus import Publisher, Consumer
from core.rpc import Caller
from module.log.component.model.activity_model import activity_model
from module.log.schema.activity import ActivityData


def register_event(
    logger: Logger,
    consumer: Consumer,
    rpc_caller: Caller,
    publisher: Publisher
):
    logger.info('🥪 Register event handlers for "log.activity"')

    @consumer.register('log_new_activity')
    async def insert(
        data: Mapping[str, Any],
    ):
        await activity_model.insert(data=ActivityData(**data))



from zrb.builtin.generator.project.create import create_project
from zrb.builtin.generator.simple_python_app.add import (
    add_simple_python_app
)
import os
import pathlib
import shutil


def test_add_simple_python_app():
    # prepare path
    dir_path = pathlib.Path(__file__).parent.absolute()
    project_path = os.path.join(dir_path, 'app')
    # remove destination if exists
    if os.path.exists(project_path):
        shutil.rmtree(project_path)

    create_project_fn = create_project.to_function()
    create_project_fn(project_dir=project_path)

    automate_path = os.path.join(project_path, '_automate')
    src_path = os.path.join(project_path, 'src')

    # first attempt should success
    first_attempt_fn = add_simple_python_app.to_function()
    first_attempt_fn(
        project_dir=project_path, app_name='simpleApp'
    )

    # _automation should exists
    assert os.path.isfile(
        os.path.join(automate_path, 'simple_app', 'local.py')
    )
    assert os.path.isfile(
        os.path.join(automate_path, 'simple_app', 'container.py')
    )
    assert os.path.isfile(
        os.path.join(automate_path, 'simple_app', 'deployment.py')
    )

    # src should exists
    assert os.path.isfile(
        os.path.join(src_path, 'simple-app', 'docker-compose.yml')
    )

    # python_app should be imported
    expected_lines = [
        'assert _project',
        'assert simple_app_local',
        'assert simple_app_container',
        'assert simple_app_image',
        'assert simple_app_deployment',
        'import _automate._project as _project',
        'import _automate.simple_app.local as simple_app_local',
        'import _automate.simple_app.container as simple_app_container',
        'import _automate.simple_app.image as simple_app_image',
        'import _automate.simple_app.deployment as simple_app_deployment',
    ]
    with open(os.path.join(project_path, 'zrb_init.py')) as f:
        content = f.read()
    for line in expected_lines:
        assert line in content

    # second attempt should fail
    is_error = False
    try:
        second_attempt_fn = add_simple_python_app.to_function()
        second_attempt_fn(
            project_dir=project_path, app_name='simpleApp'
        )
    except Exception:
        is_error = True
    assert is_error



from core.messagebus.messagebus import (
    Admin, Consumer, Publisher, MessageSerializer
)
from core.messagebus.mock import MockAdmin, MockConsumer, MockPublisher
from core.messagebus.kafka.admin import KafkaAdmin
from core.messagebus.kafka.consumer import KafkaConsumer
from core.messagebus.kafka.publisher import KafkaPublisher
from core.messagebus.rabbitmq.admin import RMQAdmin
from core.messagebus.rabbitmq.consumer import RMQConsumer
from core.messagebus.rabbitmq.publisher import RMQPublisher

assert Admin
assert Consumer
assert Publisher
assert MessageSerializer
assert MockAdmin
assert MockConsumer
assert MockPublisher
assert KafkaAdmin
assert KafkaConsumer
assert KafkaPublisher
assert RMQAdmin
assert RMQConsumer
assert RMQPublisher



from zrb.helper.typing import List, Iterable
from zrb.helper.typecheck import typechecked
import fnmatch
import glob


@typechecked
def get_file_names(
    glob_path: str, glob_ignored_paths: Iterable[str]
) -> List[str]:
    matches = []
    for file in glob.glob(glob_path, recursive=True):
        should_ignore = any(
            fnmatch.fnmatch(file, ignored_path) or
            fnmatch.fnmatch(file, ignored_path + '**')
            for ignored_path in glob_ignored_paths
        )
        if not should_ignore:
            matches.append(file)
    return matches



from typing import Any
from logging import Logger
from core.messagebus import Publisher
from core.rpc import Caller, Server


def register_rpc(
    logger: Logger,
    rpc_server: Server,
    rpc_caller: Caller,
    publisher: Publisher
):
    logger.info('🥪 Register RPC handlers for "snake_zrb_module_name"')



from zrb.config.config import logging_level
from zrb.helper.typecheck import typechecked
from zrb.helper.log import logger
from zrb.helper.accessories.color import colored
from functools import lru_cache

import os
import logging
import re

pattern = re.compile('[^a-zA-Z0-9]')


@lru_cache
@typechecked
def load_module(script_path: str):
    if not os.path.isfile(script_path):
        return
    import importlib.util
    module_name = pattern.sub('', script_path)
    if logging_level <= logging.INFO:
        logger.info(colored(f'Get module spec: {script_path}', attrs=['dark']))
    spec = importlib.util.spec_from_file_location(module_name, script_path)
    if logging_level <= logging.INFO:
        logger.info(colored(f'Create module: {script_path}', attrs=['dark']))
    module = importlib.util.module_from_spec(spec)
    if logging_level <= logging.INFO:
        logger.info(colored(f'Exec module: {script_path}', attrs=['dark']))
    spec.loader.exec_module(module)
    if logging_level <= logging.INFO:
        logger.info(colored(f'Module executed: {script_path}', attrs=['dark']))



from abc import ABC, abstractmethod


class PasswordHasher(ABC):

    @abstractmethod
    def hash_password(self, password: str) -> str:
        pass

    @abstractmethod
    def check_password(self, password: str, hashed_password: str) -> bool:
        pass



from zrb.helper.typing import Any, Callable
from zrb.helper.typecheck import typechecked
from zrb.config.config import get_current_shell
from zrb.task.task import Task
from zrb.task.cmd_task import CmdTask
from zrb.task.flow_task import FlowTask
from zrb.task_input.str_input import StrInput
from zrb.task_input.bool_input import BoolInput
from zrb.runner import runner
from zrb.helper.file.text import read_text_file_async, write_text_file_async
from zrb.builtin.group import dev_tool_install_group
import os

dir_path = os.path.dirname(__file__)
current_shell = get_current_shell()

###############################################################################
# 🔤 Input Definitions
###############################################################################

terminal_config_file_input = StrInput(
    name='config-file',
    shortcut='c',
    prompt='Config file',
    default='~/.zshrc' if current_shell == 'zsh' else '~/.bashrc'
)

###############################################################################
# Helper Definitions
###############################################################################


@typechecked
def write_config(
    template_file: str, config_file: str, remove_old_config: bool = False
) -> Callable[..., Any]:
    async def set_config(*args, **kwargs):
        task: Task = kwargs.get('_task')
        rendered_config_file = os.path.expandvars(os.path.expanduser(
            task.render_str(config_file)
        ))
        rendered_template_file = os.path.expandvars(os.path.expanduser(
            task.render_str(template_file)
        ))
        if remove_old_config and os.path.exists(rendered_config_file):
            task.print_out(f'Removing {rendered_config_file}')
            os.remove(rendered_config_file)
        additional_content = await read_text_file_async(
            rendered_template_file
        )
        content = ''
        if os.path.exists(rendered_config_file):
            content = await read_text_file_async(rendered_config_file) + '\n'
        new_content = content + additional_content
        task.print_out(f'Writing content to {rendered_config_file}')
        await write_text_file_async(rendered_config_file, new_content)
    return set_config


###############################################################################
# Task Definitions
###############################################################################

install_gvm = FlowTask(
    name='gvm',
    group=dev_tool_install_group,
    description='GVM provides interface to manage go version',
    inputs=[
        StrInput(
            name='go-default-version',
            description='Go default version',
            default='go1.21'
        ),
        terminal_config_file_input,
    ],
    steps=[
        CmdTask(
            name='download-gvm',
            cmd_path=os.path.join(dir_path, 'gvm', 'download.sh'),
            preexec_fn=None
        ),
        Task(
            name='configure-gvm',
            run=write_config(
                template_file=os.path.join(
                    dir_path, 'gvm', 'resource', 'config.sh'
                ),
                config_file='{{input.config_file}}'
            )
        ),
        CmdTask(
            name='finalize-gvm-installation',
            cmd_path=os.path.join(dir_path, 'gvm', 'finalize.sh'),
            preexec_fn=None
        )
    ],
    retry=0
)
runner.register(install_gvm)

install_pyenv = FlowTask(
    name='pyenv',
    group=dev_tool_install_group,
    description='Simple Python version management',
    inputs=[
        StrInput(
            name='python-default-version',
            description='Python default version',
            default='3.10.0'
        ),
        terminal_config_file_input,
    ],
    steps=[
        CmdTask(
            name='download-pyenv',
            cmd_path=os.path.join(dir_path, 'pyenv', 'download.sh'),
            preexec_fn=None
        ),
        Task(
            name='configure-pyenv',
            run=write_config(
                template_file=os.path.join(
                    dir_path, 'pyenv', 'resource', 'config.sh'
                ),
                config_file='{{input.config_file}}'
            )
        ),
        CmdTask(
            name='finalize-pyenv-installation',
            cmd_path=os.path.join(dir_path, 'pyenv', 'finalize.sh'),
            preexec_fn=None
        )
    ],
    retry=0
)
runner.register(install_pyenv)

install_nvm = FlowTask(
    name='nvm',
    group=dev_tool_install_group,
    description='NVM allows you to quickly install and use different versions of node via the command line', # noqa
    inputs=[
        StrInput(
            name='node-default-version',
            description='Node default version',
            default='node'
        ),
        terminal_config_file_input,
    ],
    steps=[
        CmdTask(
            name='download-nvm',
            cmd_path=os.path.join(dir_path, 'nvm', 'download.sh'),
            preexec_fn=None
        ),
        Task(
            name='configure-nvm',
            run=write_config(
                template_file=os.path.join(
                    dir_path, 'nvm', 'resource', 'config.sh'
                ),
                config_file='{{input.config_file}}'
            )
        ),
        CmdTask(
            name='finalize-nvm-installation',
            cmd_path=os.path.join(dir_path, 'nvm', 'finalize.sh'),
            preexec_fn=None
        )
    ],
    retry=0
)
runner.register(install_nvm)

install_sdkman = FlowTask(
    name='sdkman',
    group=dev_tool_install_group,
    description='SDKMAN! is a tool for managing parallel versions of multiple Software Development Kits on most Unix based systems', # noqa
    inputs=[
        BoolInput(
            name='install-java',
            description='Install Java',
            prompt='Do you want to install Java?',
            default=True
        ),
        BoolInput(
            name='install-scala',
            description='Install Scala',
            prompt='Do you want to install Scala?',
            default=True
        ),
        terminal_config_file_input,
    ],
    steps=[
        CmdTask(
            name='download-sdkman',
            cmd_path=os.path.join(dir_path, 'sdkman', 'download.sh'),
            preexec_fn=None
        ),
        Task(
            name='configure-sdkman',
            run=write_config(
                template_file=os.path.join(
                    dir_path, 'sdkman', 'resource', 'config.sh'
                ),
                config_file='{{input.config_file}}'
            )
        ),
        CmdTask(
            name='finalize-sdkman-installation',
            cmd_path=os.path.join(dir_path, 'sdkman', 'finalize.sh'),
            preexec_fn=None
        )
    ],
    retry=0
)
runner.register(install_sdkman)

install_pulumi = FlowTask(
    name='pulumi',
    group=dev_tool_install_group,
    description='Universal infrastructure as code',
    inputs=[terminal_config_file_input],
    steps=[
        CmdTask(
            name='install-pulumi',
            cmd_path=os.path.join(dir_path, 'pulumi', 'install.sh'),
            preexec_fn=None
        ),
        Task(
            name='configure-pulumi',
            run=write_config(
                template_file=os.path.join(
                    dir_path, 'pulumi', 'resource', 'config.sh'
                ),
                config_file='{{input.config_file}}'
            )
        ),
    ],
    retry=0
)
runner.register(install_pulumi)

install_aws = FlowTask(
    name='aws',
    group=dev_tool_install_group,
    description='AWS CLI',
    steps=[
        CmdTask(
            name='install-aws',
            cmd_path=os.path.join(dir_path, 'aws', 'install.sh'),
            preexec_fn=None
        ),
    ],
    retry=0
)
runner.register(install_aws)

install_gcloud = FlowTask(
    name='gcloud',
    group=dev_tool_install_group,
    description='Gcloud CLI',
    steps=[
        CmdTask(
            name='install-gcloud',
            cmd_path=os.path.join(dir_path, 'gcloud', 'install.sh'),
            preexec_fn=None
        ),
    ],
    retry=0
)
runner.register(install_gcloud)

install_tmux = FlowTask(
    name='tmux',
    group=dev_tool_install_group,
    description='Terminal multiplexer',
    inputs=[
        StrInput(
            name='config-file',
            shortcut='c',
            prompt='Config file',
            default='~/.tmux.conf'
        )
    ],
    steps=[
        CmdTask(
            name='install-tmux',
            cmd_path=os.path.join(dir_path, 'tmux', 'install.sh'),
            preexec_fn=None
        ),
        Task(
            name='configure-tmux',
            run=write_config(
                template_file=os.path.join(
                    dir_path, 'tmux', 'resource', 'config.sh'
                ),
                config_file='{{input.config_file}}'
            )
        ),
    ],
    retry=0
)
runner.register(install_tmux)

install_zsh = FlowTask(
    name='zsh',
    group=dev_tool_install_group,
    description='Zsh terminal + oh-my-zsh + zdharma',
    inputs=[
        StrInput(
            name='config-file',
            shortcut='c',
            prompt='Config file',
            default='~/.zshrc'
        )
    ],
    steps=[
        CmdTask(
            name='install-zsh',
            cmd_path=os.path.join(dir_path, 'zsh', 'install.sh'),
            preexec_fn=None
        ),
        Task(
            name='configure-zsh',
            run=write_config(
                template_file=os.path.join(
                    dir_path, 'zsh', 'resource', 'config.sh'
                ),
                config_file='{{input.config_file}}'
            )
        ),
    ],
    retry=0
)
runner.register(install_zsh)

install_kubectl = FlowTask(
    name='kubectl',
    group=dev_tool_install_group,
    description='Kubernetes CLI tool',
    steps=[
        CmdTask(
            name='install-kubectl',
            cmd_path=os.path.join(dir_path, 'kubectl', 'install.sh'),
            preexec_fn=None
        ),
    ],
    retry=0
)
runner.register(install_kubectl)

install_helm = FlowTask(
    name='helm',
    group=dev_tool_install_group,
    description='Package manager for kubernetes',
    steps=[
        CmdTask(
            name='install-helm',
            cmd_path=os.path.join(dir_path, 'helm', 'install.sh'),
            preexec_fn=None
        ),
    ],
    retry=0
)
runner.register(install_helm)

install_docker = FlowTask(
    name='docker',
    group=dev_tool_install_group,
    description='Most popular containerization platform',
    steps=[
        CmdTask(
            name='install-docker',
            cmd_path=os.path.join(dir_path, 'docker', 'install.sh'),
            preexec_fn=None
        ),
    ],
    retry=0
)
runner.register(install_docker)

install_terraform = FlowTask(
    name='terraform',
    group=dev_tool_install_group,
    description='Open source IAC by Hashicorp',
    inputs=[
        terminal_config_file_input
    ],
    steps=[
        CmdTask(
            name='install-terraform',
            cmd_path=os.path.join(dir_path, 'terraform', 'install.sh'),
            preexec_fn=None
        ),
        Task(
            name='configure-terraform',
            run=write_config(
                template_file=os.path.join(
                    dir_path, 'terraform', 'resource', 'config.sh'
                ),
                config_file='{{input.config_file}}'
            )
        ),
    ],
    retry=0
)
runner.register(install_terraform)

install_helix = FlowTask(
    name='helix',
    group=dev_tool_install_group,
    description='Post modern text editor',
    steps=[
        CmdTask(
            name='install-helix',
            cmd_path=os.path.join(dir_path, 'helix', 'install.sh'),
            preexec_fn=None
        ),
        [
            Task(
                name='create-helix-theme',
                run=write_config(
                    template_file=os.path.join(
                        dir_path, 'helix', 'resource', 'themes', 'gruvbox_transparent.toml'  # noqa
                    ),
                    config_file='~/.config/helix/themes/gruvbox_transparent.toml',  # noqa
                    remove_old_config=True
                )
            ),
            Task(
                name='configure-helix',
                run=write_config(
                    template_file=os.path.join(
                        dir_path, 'helix', 'resource', 'config.toml'
                    ),
                    config_file='~/.config/helix/config.toml',
                    remove_old_config=True
                )
            ),
            CmdTask(
                name='install-language-server',
                cmd_path=os.path.join(
                    dir_path, 'helix', 'install-language-server.sh'
                ),
                preexec_fn=None
            ),
        ],
    ],
    retry=0
)
runner.register(install_helix)

install_selenium = FlowTask(
    name='selenium',
    group=dev_tool_install_group,
    description='Selenium + Chrome web driver',
    steps=[
        CmdTask(
            name='install-selenium',
            cmd_path=os.path.join(dir_path, 'selenium', 'install.sh'),
            preexec_fn=None
        )
    ]
)
runner.register(install_selenium)



from zrb.helper.typing import Optional
from zrb.helper.typecheck import typechecked

import asyncio
import time

LOG_NAME_LENGTH = 20


@typechecked
class TimeTracker():

    def __init__(self):
        self.__start_time: float = 0
        self.__end_time: float = 0

    def _start_timer(self):
        self.__start_time = time.time()

    def _end_timer(self):
        self.__end_time = time.time()

    def _get_elapsed_time(self) -> float:
        return self.__end_time - self.__start_time


@typechecked
class AttemptTracker():

    def __init__(self, retry: int = 2):
        self.__retry = retry
        self.__attempt: int = 1

    def _get_max_attempt(self) -> int:
        return self.__retry + 1

    def _get_attempt(self) -> int:
        return self.__attempt

    def _increase_attempt(self):
        self.__attempt += 1

    def _should_attempt(self) -> bool:
        attempt = self._get_attempt()
        max_attempt = self._get_max_attempt()
        return attempt <= max_attempt

    def _is_last_attempt(self) -> bool:
        attempt = self._get_attempt()
        max_attempt = self._get_max_attempt()
        return attempt >= max_attempt


@typechecked
class FinishTracker():

    def __init__(self):
        self.__execution_queue: Optional[asyncio.Queue] = None
        self.__counter = 0

    async def _mark_awaited(self):
        if self.__execution_queue is None:
            self.__execution_queue = asyncio.Queue()
        self.__counter += 1

    async def _mark_done(self):
        # Tracker might be started several times
        # However, when the execution is marked as done, it applied globally
        # Thus, we need to send event as much as the counter.
        for i in range(self.__counter):
            await self.__execution_queue.put(True)

    async def _is_done(self) -> bool:
        while self.__execution_queue is None:
            await asyncio.sleep(0.05)
        return await self.__execution_queue.get()



from zrb.task.any_task import AnyTask
from zrb.helper.typing import Any, JinjaTemplate, Mapping, Optional, Union
from zrb.helper.typecheck import typechecked
from zrb.helper.string.conversion import to_boolean
from zrb.helper.string.jinja import is_probably_jinja
from zrb.helper.render_data import DEFAULT_RENDER_DATA

import os
import jinja2


class AnyExtensionFileSystemLoader(jinja2.FileSystemLoader):
    def get_source(self, environment, template):
        for search_dir in self.searchpath:
            file_path = os.path.join(search_dir, template)
            if os.path.exists(file_path):
                with open(file_path, 'r') as file:
                    contents = file.read()
                return contents, file_path, lambda: False
        raise jinja2.TemplateNotFound(template)


@typechecked
class Renderer():

    def __init__(self):
        self.__input_map: Mapping[str, Any] = {}
        self.__task: Optional[AnyTask] = None
        self.__env_map: Mapping[str, str] = {}
        self.__render_data: Optional[Mapping[str, Any]] = None
        self.__rendered_str: Mapping[str, str] = {}

    def _set_task(self, task: AnyTask):
        self.__task = task

    def get_input_map(self) -> Mapping[str, Any]:
        # This return reference to input map, so input map can be updated
        return self.__input_map

    def _set_input_map(self, key: str, val: Any):
        self.__input_map[key] = val

    def get_env_map(self) -> Mapping[str, str]:
        # This return reference to env map, so env map can be updated
        return self.__env_map

    def _set_env_map(self, key: str, val: str):
        self.__env_map[key] = val

    def render_any(
        self, value: Any, data: Optional[Mapping[str, Any]] = None
    ) -> Any:
        if isinstance(value, str):
            return self.render_str(value, data)
        return value

    def render_float(
        self,
        value: Union[JinjaTemplate, float],
        data: Optional[Mapping[str, Any]] = None
    ) -> float:
        if isinstance(value, str):
            return float(self.render_str(value, data))
        return value

    def render_int(
        self,
        value: Union[JinjaTemplate, int],
        data: Optional[Mapping[str, Any]] = None
    ) -> int:
        if isinstance(value, str):
            return int(self.render_str(value, data))
        return value

    def render_bool(
        self,
        value: Union[JinjaTemplate, bool],
        data: Optional[Mapping[str, Any]] = None
    ) -> bool:
        if isinstance(value, str):
            return to_boolean(self.render_str(value, data))
        return value

    def render_str(
        self,
        value: JinjaTemplate,
        data: Optional[Mapping[str, Any]] = None
    ) -> str:
        if value in self.__rendered_str:
            return self.__rendered_str[value]
        if not is_probably_jinja(value):
            return value
        template = jinja2.Template(value)
        render_data = self.__get_render_data(additional_data=data)
        try:
            rendered_text = template.render(render_data)
        except Exception:
            raise Exception(
                f'Fail to render "{value}" with data: {render_data}'
            )
        self.__rendered_str[value] = rendered_text
        return rendered_text

    def render_file(
        self,
        path: JinjaTemplate,
        data: Optional[Mapping[str, Any]] = None
    ) -> str:
        location_dir = os.path.dirname(path)
        env = jinja2.Environment(
            loader=AnyExtensionFileSystemLoader([location_dir])
        )
        template = env.get_template(path)
        render_data = self.__get_render_data(additional_data=data)
        render_data['TEMPLATE_DIR'] = location_dir
        rendered_text = template.render(render_data)
        return rendered_text

    def __get_render_data(
        self, additional_data: Optional[Mapping[str, Any]] = None
    ) -> Mapping[str, Any]:
        self.__ensure_cached_render_data()
        if additional_data is None:
            return self.__render_data
        return {**self.__render_data, **additional_data}

    def __ensure_cached_render_data(self):
        if self.__render_data is not None:
            return self.__render_data
        render_data = dict(DEFAULT_RENDER_DATA)
        render_data.update({
            'env': self.__env_map,
            'input': self.__input_map,
            'task': self.__task,
        })
        self.__render_data = render_data
        return render_data



from pulumi_kubernetes.helm.v3 import Chart, LocalChartOpts
from _common import (
    NAMESPACE, SIGNOZ_CLICKHOUSE_NAMESPACE, SIGNOZ_CLICKHOUSE_USER,
    SIGNOZ_CLICKHOUSE_PASSWORD
)


def create_signoz() -> Chart:
    signoz = Chart(
        'signoz',
        LocalChartOpts(
            path='./helm-charts/signoz',
            namespace=NAMESPACE,
            values={
                'clickhouse': {
                    'enabled': True,
                    'user': SIGNOZ_CLICKHOUSE_USER,
                    'password': SIGNOZ_CLICKHOUSE_PASSWORD,
                    'namespace': SIGNOZ_CLICKHOUSE_NAMESPACE,
                },
                'resources': {
                    'limits': {
                        'cpu': '250m',
                        'memory': '4Gi'
                    },
                    'requests': {
                        'cpu': '250m',
                        'memory': '200Mi'
                    },
                }
            },
            skip_await=True
        ),
    )
    return signoz



from zrb import HTTPChecker, PortChecker
from ._helper import should_start_local_microservices

rabbitmq_management_checker = HTTPChecker(
    name='check-rabbitmq-management',
    port='{{env.get("RABBITMQ_MANAGEMENT_HOST_PORT", "15672")}}',
    is_https='{{input.snake_zrb_app_name_https}}',
    should_execute='{{env.get("APP_BROKER_TYPE", "rabbitmq") == "rabbitmq"}}'
)

rabbitmq_checker = PortChecker(
    name='check-rabbitmq',
    port='{{env.get("RABBITMQ_HOST_PORT", "5672")}}',
    should_execute='{{env.get("APP_BROKER_TYPE", "rabbitmq") == "rabbitmq"}}'
)

redpanda_console_checker = HTTPChecker(
    name='check-redpanda-console',
    method='GET',
    port='{{env.get("REDPANDA_CONSOLE_HOST_PORT", "9000")}}',
    is_https='{{input.snake_zrb_app_name_https}}',
    should_execute='{{env.get("APP_BROKER_TYPE", "rabbitmq") == "kafka"}}'
)

kafka_plaintext_checker = PortChecker(
    name='check-kafka-plaintext',
    port='{{env.get("KAFKA_INTERNAL_HOST_PORT", "29092")}}',
    should_execute='{{env.get("APP_BROKER_TYPE", "rabbitmq") == "kafka"}}'
)

kafka_outside_checker = PortChecker(
    name='check-kafka-outside',
    port='{{env.get("KAFKA_EXTERNAL_HOST_PORT", "9092")}}',
    should_execute='{{env.get("APP_BROKER_TYPE", "rabbitmq") == "kafka"}}'
)

pandaproxy_plaintext_checker = PortChecker(
    name='check-pandaproxy-plaintext',
    port='{{env.get("PANDAPROXY_INTERNAL_HOST_PORT", "29092")}}',
    should_execute='{{env.get("APP_BROKER_TYPE", "rabbitmq") == "kafka"}}'
)

pandaproxy_outside_checker = PortChecker(
    name='check-pandaproxy-outside',
    port='{{env.get("PANDAPROXY_EXTERNAL_HOST_PORT", "9092")}}',
    should_execute='{{env.get("APP_BROKER_TYPE", "rabbitmq") == "kafka"}}'
)

app_container_checker = HTTPChecker(
    name='check-kebab-zrb-app-name-container',
    host='{{input.snake_zrb_app_name_host}}',
    url='/readiness',
    port='{{env.get("HOST_PORT", "zrbAppHttpPort")}}',
    is_https='{{input.snake_zrb_app_name_https}}'
)

app_local_checker = HTTPChecker(
    name='check-kebab-zrb-app-name',
    host='{{input.snake_zrb_app_name_host}}',
    url='/readiness',
    port='{{env.APP_PORT}}',
    is_https='{{input.snake_zrb_app_name_https}}',
    should_execute=should_start_local_microservices
)



from zrb.builtin.generator.project.create import create_project
from zrb.config.config import version
import os
import pathlib
import shutil


def test_create_project():
    # prepare path
    dir_path = pathlib.Path(__file__).parent.absolute()
    destination_path = os.path.join(dir_path, 'app')
    # remove destination if exists
    if os.path.exists(destination_path):
        shutil.rmtree(destination_path)

    # first attempt should success
    first_attempt_fn = create_project.to_function()
    result = first_attempt_fn(project_dir=destination_path)
    assert result

    # .git directory should exists
    assert os.path.isdir(os.path.join(destination_path, '.git'))
    # .gitignore file should exists
    assert os.path.isfile(os.path.join(destination_path, '.gitignore'))
    # zrb_init.py should exists
    assert os.path.isfile(os.path.join(destination_path, 'zrb_init.py'))

    with open(os.path.join(destination_path, 'README.md')) as readme_file:
        readme_lines = readme_file.readlines()
    assert readme_lines[0] == '# App\n'

    with open(
        os.path.join(destination_path, 'requirements.txt')
    ) as requirements_file:
        requirements_lines = requirements_file.readlines()
    assert requirements_lines[0] == f'zrb=={version}\n'

    # second attempt should failed
    is_error = False
    try:
        second_attempt_fn = create_project.to_function()
        result = second_attempt_fn(project_dir=destination_path)
    except Exception:
        is_error = True
    assert is_error



from zrb import BoolInput, StrInput
import os

###############################################################################
# 🏔️ Constants
###############################################################################

CURRENT_DIR = os.path.dirname(__file__)
PROJECT_DIR = os.path.abspath(os.path.join(CURRENT_DIR, '..', '..'))
RESOURCE_DIR = os.path.join(PROJECT_DIR, 'src', 'kebab-zrb-app-name')
DEPLOYMENT_DIR = os.path.join(RESOURCE_DIR, 'deployment')
DEPLOYMENT_TEMPLATE_ENV_FILE_NAME = os.path.join(
    DEPLOYMENT_DIR, 'template.env'
)
APP_DIR = os.path.join(RESOURCE_DIR, 'src')
APP_TEMPLATE_ENV_FILE_NAME = os.path.join(APP_DIR, 'template.env')

###############################################################################
# 🔤 Input Definitions
###############################################################################

local_input = BoolInput(
    name='local-kebab-zrb-app-name',
    description='Use "kebab-zrb-app-name" on local machine',
    prompt='Use "kebab-zrb-app-name" on local machine?',
    default=True
)

https_input = BoolInput(
    name='kebab-zrb-app-name-https',
    description='Whether "kebab-zrb-app-name" run on HTTPS',
    prompt='Is "kebab-zrb-app-name" run on HTTPS?',
    default=False
)

host_input = StrInput(
    name='kebab-zrb-app-name-host',
    description='Hostname of "kebab-zrb-app-name"',
    prompt='Hostname of "kebab-zrb-app-name"',
    default='localhost'
)



from zrb.helper.typing import (
    Any, Callable, Iterable, List, Mapping, Optional, Union
)
from zrb.helper.typecheck import typechecked
from zrb.config.config import show_time, logging_level
from zrb.task.any_task import AnyTask
from zrb.task.any_task_event_handler import (
    OnTriggered, OnWaiting, OnSkipped, OnStarted, OnReady, OnRetry, OnFailed
)
from zrb.helper.log import logger
from zrb.helper.accessories.color import colored
from zrb.task_input.any_input import AnyInput
from zrb.task_group.group import Group
from zrb.task_env.env import Env
from zrb.task_env.env_file import EnvFile
from zrb.task.base_task.component.common_task_model import CommonTaskModel
from zrb.task.base_task.component.pid_model import PidModel
from zrb.task.base_task.component.trackers import TimeTracker
from zrb.config.config import env_prefix
from zrb.helper.string.modification import double_quote
from zrb.helper.string.conversion import to_variable_name
from functools import lru_cache

import datetime
import logging
import os
import sys

LOG_NAME_LENGTH = 20


@typechecked
class BaseTaskModel(CommonTaskModel, PidModel, TimeTracker):
    def __init__(
        self,
        name: str,
        group: Optional[Group] = None,
        description: str = '',
        inputs: List[AnyInput] = [],
        envs: Iterable[Env] = [],
        env_files: Iterable[EnvFile] = [],
        icon: Optional[str] = None,
        color: Optional[str] = None,
        retry: int = 2,
        retry_interval: Union[int, float] = 1,
        upstreams: Iterable[AnyTask] = [],
        checkers: Iterable[AnyTask] = [],
        checking_interval: Union[int, float] = 0,
        run: Optional[Callable[..., Any]] = None,
        on_triggered: Optional[OnTriggered] = None,
        on_waiting: Optional[OnWaiting] = None,
        on_skipped: Optional[OnSkipped] = None,
        on_started: Optional[OnStarted] = None,
        on_ready: Optional[OnReady] = None,
        on_retry: Optional[OnRetry] = None,
        on_failed: Optional[OnFailed] = None,
        should_execute: Union[bool, str, Callable[..., bool]] = True,
        return_upstream_result: bool = False
    ):
        self.__rjust_full_cli_name: Optional[str] = None
        self.__has_cli_interface = False
        self.__complete_name: Optional[str] = None
        CommonTaskModel.__init__(
            self,
            name=name,
            group=group,
            description=description,
            inputs=inputs,
            envs=envs,
            env_files=env_files,
            icon=icon,
            color=color,
            retry=retry,
            retry_interval=retry_interval,
            upstreams=upstreams,
            checkers=checkers,
            checking_interval=checking_interval,
            run=run,
            on_triggered=on_triggered,
            on_waiting=on_waiting,
            on_skipped=on_skipped,
            on_started=on_started,
            on_ready=on_ready,
            on_retry=on_retry,
            on_failed=on_failed,
            should_execute=should_execute,
            return_upstream_result=return_upstream_result
        )
        PidModel.__init__(self)
        TimeTracker.__init__(self)
        self.__args: List[Any] = []
        self.__kwargs: Mapping[str, Any] = {}

    def _set_args(self, args: Iterable[Any]):
        '''
        Set args that will be shown at the end of the execution
        '''
        self.__args = list(args)

    def _set_kwargs(self, kwargs: Mapping[str, Any]):
        '''
        Set kwargs that will be shown at the end of the execution
        '''
        self.__kwargs = kwargs

    def log_debug(self, message: Any):
        if logging_level > logging.DEBUG:
            return
        prefix = self.__get_log_prefix()
        colored_message = colored(
            f'{prefix} • {message}', attrs=['dark']
        )
        logger.debug(colored_message)

    def log_warn(self, message: Any):
        if logging_level > logging.WARNING:
            return
        prefix = self.__get_log_prefix()
        colored_message = colored(
            f'{prefix} • {message}', attrs=['dark']
        )
        logger.warning(colored_message)

    def log_info(self, message: Any):
        if logging_level > logging.INFO:
            return
        prefix = self.__get_log_prefix()
        colored_message = colored(
            f'{prefix} • {message}', attrs=['dark']
        )
        logger.info(colored_message)

    def log_error(self, message: Any):
        if logging_level > logging.ERROR:
            return
        prefix = self.__get_log_prefix()
        colored_message = colored(
            f'{prefix} • {message}', color='red', attrs=['bold']
        )
        logger.error(colored_message, exc_info=True)

    def log_critical(self, message: Any):
        if logging_level > logging.CRITICAL:
            return
        prefix = self.__get_log_prefix()
        colored_message = colored(
            f'{prefix} • {message}', color='red', attrs=['bold']
        )
        logger.critical(colored_message, exc_info=True)

    def print_out(self, message: Any, trim_message: bool = True):
        prefix = self.__get_colored_print_prefix()
        message_str = f'{message}'.rstrip() if trim_message else f'{message}'
        print(f'🤖 ○ {prefix} • {message_str}', file=sys.stderr)
        sys.stderr.flush()

    def print_err(self, message: Any, trim_message: bool = True):
        prefix = self.__get_colored_print_prefix()
        message_str = f'{message}'.rstrip() if trim_message else f'{message}'
        print(f'🤖 △ {prefix} • {message_str}', file=sys.stderr)
        sys.stderr.flush()

    def print_out_dark(self, message: Any, trim_message: bool = True):
        message_str = f'{message}'
        self.print_out(colored(message_str, attrs=['dark']), trim_message)

    def _print_result(self, result: Any):
        if result is None:
            return
        if self._return_upstream_result:
            # if _return_upstream_result, result is list (see: self._run_all)
            upstreams = self._get_upstreams()
            upstream_results = list(result)
            for upstream_index, upstream_result in enumerate(upstream_results):
                upstreams[upstream_index]._print_result(upstream_result)
            return
        self.print_result(result)

    def print_result(self, result: Any):
        print(result)

    def _play_bell(self):
        print('\a', end='', file=sys.stderr, flush=True)

    def _show_done_info(self):
        elapsed_time = self._get_elapsed_time()
        self.print_out_dark(f'Completed in {elapsed_time} seconds')
        self._play_bell()

    def _show_env_prefix(self):
        if env_prefix == '':
            return
        colored_env_prefix = colored(env_prefix, color='yellow')
        colored_label = colored('Your current environment: ', attrs=['dark'])
        print(colored(f'{colored_label}{colored_env_prefix}'), file=sys.stderr)

    def _show_run_command(self):
        params: List[str] = [double_quote(arg) for arg in self.__args]
        for task_input in self._get_combined_inputs():
            if task_input.is_hidden():
                continue
            key = task_input.get_name()
            kwarg_key = to_variable_name(key)
            quoted_value = double_quote(str(self.__kwargs[kwarg_key]))
            params.append(f'--{key} {quoted_value}')
        run_cmd = self._get_full_cli_name()
        run_cmd_with_param = run_cmd
        if len(params) > 0:
            param_str = ' '.join(params)
            run_cmd_with_param += ' ' + param_str
        colored_command = colored(run_cmd_with_param, color='yellow')
        colored_label = colored('To run again: ', attrs=['dark'])
        print(colored(f'{colored_label}{colored_command}'), file=sys.stderr)

    def __get_colored_print_prefix(self) -> str:
        return self.__get_colored(self.__get_print_prefix())

    def __get_colored(self, text: str) -> str:
        return colored(text, color=self.get_color())

    def __get_print_prefix(self) -> str:
        common_prefix = self.__get_common_prefix(show_time=show_time)
        icon = self.get_icon()
        length = LOG_NAME_LENGTH - len(icon)
        rjust_cli_name = self.__get_rjust_full_cli_name(length)
        return f'{common_prefix} {icon} {rjust_cli_name}'

    def __get_log_prefix(self) -> str:
        common_prefix = self.__get_common_prefix(show_time=False)
        icon = self.get_icon()
        length = LOG_NAME_LENGTH - len(icon)
        filled_name = self.__get_rjust_full_cli_name(length)
        return f'{common_prefix} {icon} {filled_name}'

    def __get_common_prefix(self, show_time: bool) -> str:
        attempt = self._get_attempt()
        max_attempt = self._get_max_attempt()
        pid = str(self._get_task_pid()).rjust(6)
        if show_time:
            now = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S.%f')[:-3]
            return f'◷ {now} ❁ {pid} → {attempt}/{max_attempt}'
        return f'❁ {pid} → {attempt}/{max_attempt}'

    @lru_cache
    def __get_rjust_full_cli_name(self, length: int) -> str:
        if self.__rjust_full_cli_name is not None:
            return self.__rjust_full_cli_name
        complete_name = self._get_full_cli_name()
        self.__rjust_full_cli_name = complete_name.rjust(length, ' ')
        return self.__rjust_full_cli_name

    @lru_cache
    def __get_executable_name(self) -> str:
        if len(sys.argv) > 0 and sys.argv[0] != '':
            return os.path.basename(sys.argv[0])
        return 'zrb'

    @lru_cache
    def _get_full_cli_name(self) -> str:
        if self.__complete_name is not None:
            return self.__complete_name
        executable_prefix = ''
        if self.__has_cli_interface:
            executable_prefix += self.__get_executable_name() + ' '
        cli_name = self.get_cli_name()
        if self._group is None:
            self.__complete_name = f'{executable_prefix}{cli_name}'
            return self.__complete_name
        group_cli_name = self._group._get_full_cli_name()
        self.__complete_name = f'{executable_prefix}{group_cli_name} {cli_name}'  # noqa
        return self.__complete_name

    def _set_has_cli_interface(self):
        self.__has_cli_interface = True



from zrb.helper.docker_compose.fetch_external_env import (
    fetch_compose_file_env_map
)


def test_fetch_external_env():
    data = {
        'services': {
            'service_1': {
                'environment': [
                    'PORT=${PORT_A:-8080}',
                    'URL=${PROTOCOL_A:-http}://${HOST_A1:-${HOST_A2:-${HOST_A3:-localhost}}}${URL_A}', # noqa
                    'SECRET=${SECRET_A}',
                    'LITERAL=LITERAL',
                ]
            },
            'service_2': {
                'environment': {
                    'PORT': '${PORT_B:-8080}',
                    'URL': '${PROTOCOL_B:-http}://${HOST_B1:-${HOST_B2:-${HOST_B3:-localhost}}}${URL_B}', # noqa
                    'SECRET': '${SECRET_A}',
                    'LITERAL': 'LITERAL',
                }
            }
        }
    }
    external_env = fetch_compose_file_env_map(data)
    assert external_env.get('PORT_A') == '8080'
    assert external_env.get('PROTOCOL_A') == 'http'
    assert external_env.get('HOST_A1') == ''
    assert external_env.get('HOST_A2') == ''
    assert external_env.get('HOST_A3') == 'localhost'
    assert external_env.get('URL_A') == ''
    assert external_env.get('PORT_B') == '8080'
    assert external_env.get('PROTOCOL_B') == 'http'
    assert external_env.get('HOST_B1') == ''
    assert external_env.get('HOST_B2') == ''
    assert external_env.get('HOST_B3') == 'localhost'
    assert external_env.get('URL_B') == ''



from zrb.helper.typecheck import typechecked
from zrb.helper.typing import List, Mapping
from zrb.task_group.group import Group
from zrb.task.any_task import AnyTask
from zrb.task_env.env import Env
from zrb.helper.string.jinja import is_probably_jinja


@typechecked
def fetch_env_map_from_group(
    env_map: Mapping[str, str], group: Group
) -> Mapping[str, str]:
    for task in group.get_tasks():
        env_map = fetch_env_map_from_task(env_map, task)
    for sub_group in group.get_children():
        sub_env_map: Mapping[str, str] = fetch_env_map_from_group(
            env_map, sub_group
        )
        env_map = _cascade_env_map(env_map, sub_env_map)
    return env_map


@typechecked
def fetch_env_map_from_task(
    env_map: Mapping[str, str], task: AnyTask
):
    task_env_map: Mapping[str, str] = {}
    for env_file in task._get_env_files():
        envs = env_file.get_envs()
        task_env_map = _add_envs_to_env_map(task_env_map, envs)
    task_env_map = _add_envs_to_env_map(task_env_map, task._envs)
    env_map = _cascade_env_map(env_map, task_env_map)
    for upstream in task._get_upstreams():
        task_env_map = fetch_env_map_from_task(env_map, upstream)
    for checker in task._get_checkers():
        task_env_map = fetch_env_map_from_task(env_map, checker)
    return env_map


@typechecked
def _add_envs_to_env_map(
    env_map: Mapping[str, str], envs: List[Env]
) -> Mapping[str, str]:
    for env in envs:
        if env.get_os_name() == '':
            continue
        env_name = _get_env_name(env)
        env_default = _get_env_default(env)
        env_map[env_name] = env_default
    return env_map


@typechecked
def _cascade_env_map(
    env_map: Mapping[str, str],
    other_env_map: Mapping[str, str]
) -> Mapping[str, str]:
    for key, value in other_env_map.items():
        if key in env_map:
            continue
        env_map[key] = value
    return env_map


@typechecked
def _get_env_name(env: Env) -> str:
    if env.get_os_name() is None:
        return env.get_name()
    return env.get_os_name()


@typechecked
def _get_env_default(env: Env) -> str:
    if is_probably_jinja(env.get_default()):
        return ''
    return env.get_default()



from zrb.task.recurring_task import RecurringTask
from zrb.task.path_watcher import PathWatcher
from zrb.task_input.str_input import StrInput
from zrb.runner import runner
from zrb.builtin.helper.reccuring_action import create_recurring_action


watch_changes = RecurringTask(
    name='watch-changes',
    icon='🕵️',
    color='yellow',
    description='Watch changes and show message/run command',
    inputs=[
        StrInput(
            name='pattern',
            default='*.*',
            prompt='File pattern',
            description='File pattern to be watched'
        ),
        StrInput(
            name='ignored-pattern',
            default='',
            prompt='Ignored file pattern',
            description='Ignored file pattern'
        ),
    ],
    triggers=[
        PathWatcher(
            name='watch-path',
            color='cyan',
            icon='👀',
            path='{{input.pattern}}',
            ignored_path='{{input.ignored_pattern}}'
        )
    ],
    task=create_recurring_action(
        notif_title='Watch',
        trigger_caption='File changes',
        trigger_xcom_key='watch-path.file'
    )
)
runner.register(watch_changes)
